# -*- coding: utf-8 -*-
"""
Tests for the pw input plugin.

TODO: to test:
- association species->pseudos
- two pseudos with the same filename
- IFPOS (FIXED_COORDS in SETTINGS)
- automatic namelists
- manually specified namelists
- empty namelists
- content for non-existent namelists specified
"""
import os

from aiida.backends.djsite.db.testbase import AiidaTestCase
from aiida.orm import CalculationFactory, DataFactory
from aiida.orm.code import Code
from aiida.common.folders import SandboxFolder
from aiida.common.exceptions import InputValidationError
import aiida

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

QECalc = CalculationFactory('quantumespresso.pw')
StructureData = DataFactory('structure')
ParameterData = DataFactory('parameter')
UpfData = DataFactory('upf')
KpointsData = DataFactory('array.kpoints')


class QETestCase(AiidaTestCase):
    @classmethod
    def setUpClass(cls):
        super(QETestCase, cls).setUpClass()
        cls.calc_params = {
            'computer': cls.computer,
            'resources': {
                'num_machines': 1,
                'num_mpiprocs_per_machine': 1}
        }
        cls.code = Code(remote_computer_exec=(cls.computer, '/x.x')).store()



class TestQEPWInputGeneration(QETestCase):
    """
    Test if the input is correctly generated
    """

    def test_inputs(self):
        import logging

        cell = ((2., 0., 0.), (0., 2., 0.), (0., 0., 2.))

        input_params = {
            'CONTROL': {
                'calculation': 'vc-relax',
                'restart_mode': 'from_scratch',
                'wf_collect': True,
            },
            'SYSTEM': {
                'ecutwfc': 47.,
                'ecutrho': 568.,
            },
            'ELECTRONS': {
                'conv_thr': 1.e-10,
            },
        }

        c = QECalc(**self.calc_params).store()
        s = StructureData(cell=cell)
        s.append_atom(position=(0., 0., 0.), symbols=['Ba'])
        s.store()

        p = ParameterData(dict=input_params).store()

        k = KpointsData()
        k.set_kpoints_mesh([4, 4, 4])
        k.store()

        pseudo_dir = os.path.join(os.path.split(aiida.__file__)[0],
                                  os.pardir, 'examples',
                                  'testdata', 'qepseudos')

        raw_pseudos = [
            ("Ba.pbesol-spn-rrkjus_psl.0.2.3-tot-pslib030.UPF", 'Ba'),
            ("Ti.pbesol-spn-rrkjus_psl.0.2.3-tot-pslib030.UPF", 'Ti'),
            ("O.pbesol-n-rrkjus_psl.0.1-tested-pslib030.UPF", 'O'),
        ]

        pseudos = {}
        # suppress debug messages
        logging.disable(logging.ERROR)
        for fname, elem in raw_pseudos:
            absname = os.path.realpath(os.path.join(pseudo_dir, fname))
            pseudo, _ = UpfData.get_or_create(
                absname, use_first=True)
            pseudos[elem] = pseudo
        # Reset logging level
        logging.disable(logging.NOTSET)

        inputdict = c.get_inputs_dict()
        inputdict.pop('code',None)

        with SandboxFolder() as f:
            # I use the same SandboxFolder more than once because nothing
            # should be written for these failing tests

            # Missing required input nodes
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)
            c.use_parameters(p)
            inputdict = c.get_inputs_dict()
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)
            c.use_structure(s)
            inputdict = c.get_inputs_dict()
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)
            c.use_kpoints(k)
            inputdict = c.get_inputs_dict()
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)
            c.use_pseudo(pseudos['Ba'], 'Ba')

            inputdict = c.get_inputs_dict()
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)
            c.use_code(self.code)
            inputdict = c.get_inputs_dict()
            c._prepare_for_submission(f, inputdict)

            # TODO: split this test in more than one
            c.use_pseudo(pseudos['Ti'], 'Ti')
            inputdict = c.get_inputs_dict()
            # Too many pseudos
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)

    def test_inputs_with_multiple_species(self):
        """
        Test the creation of the input file when there are two species
        associated to the same element, with different starting_magnetization
        values.
        """
        import logging

        s = StructureData(cell=[
            [2.871, 0., 0.],
            [0., 2.871, 0.],
            [0., 0., 2.871]])

        ## I leave this as a reference, but I use instead the
        ## append_atom method
        # from aiida.orm.data.structure import Kind, Site
        #s.append_kind(Kind(symbols='Ba', name='Ba1'))
        #s.append_kind(Kind(symbols='Ba', name='Ba2'))
        #s.append_site(Site(kind_name='Ba1', position=[0.,0.,0.]))
        #s.append_site(Site(kind_name='Ba2', position=[1.4355,1.4355,1.4355]))
        s.append_atom(symbols='Ba', position=[0., 0., 0.], name='Ba1')
        s.append_atom(symbols='Ba', position=[1.4355, 1.4355, 1.4355], name='Ba2')

        input_params = {
            'CONTROL': {
                'calculation': 'vc-relax',
                'restart_mode': 'from_scratch',
                'wf_collect': True,
            },
            'SYSTEM': {
                'ecutwfc': 47.,
                'ecutrho': 568.,
                'nspin': 2,
                'starting_magnetization': {'Ba1': 0.5,
                                           'Ba2': -0.5},
            },
            'ELECTRONS': {
                'conv_thr': 1.e-10,
            },
        }

        c = QECalc(**self.calc_params).store()
        c.use_code(self.code)

        p = ParameterData(dict=input_params).store()

        k = KpointsData()
        k.set_kpoints_mesh([4, 4, 4])
        k.store()

        pseudo_dir = os.path.join(os.path.split(aiida.__file__)[0],
                                  os.pardir, 'examples',
                                  'testdata', 'qepseudos')

        raw_pseudos = [
            ("Ba.pbesol-spn-rrkjus_psl.0.2.3-tot-pslib030.UPF", 'Ba'),
        ]

        pseudos = {}
        # suppress debug messages
        logging.disable(logging.ERROR)
        for fname, elem in raw_pseudos:
            absname = os.path.realpath(os.path.join(pseudo_dir, fname))
            pseudo, _ = UpfData.get_or_create(
                absname, use_first=True)
            pseudos[elem] = pseudo
        # Reset logging level
        logging.disable(logging.NOTSET)

        c.use_parameters(p)
        c.use_structure(s)
        c.use_kpoints(k)

        with SandboxFolder() as f:
            # I use the same SandboxFolder more than once because nothing
            # should be written for these failing tests

            # Same pseudo for two species
            c.use_pseudo(pseudos['Ba'], ['Ba1', 'Ba2'])
            inputdict = c.get_inputs_dict()
            c._prepare_for_submission(f, inputdict)

            with open(os.path.join(f.abspath, 'aiida.in')) as infile:
                lines = [_.strip() for _ in infile.readlines()]

            find_kind_ba1 = any('Ba1' in l and
                                'Ba.pbesol-spn-rrkjus_psl.0.2.3-tot-pslib030.UPF' in l
                                for l in lines)
            self.assertTrue(find_kind_ba1, "Unable to find the species line "
                                           "for Ba1")
            find_kind_ba2 = any('Ba2' in l and
                                'Ba.pbesol-spn-rrkjus_psl.0.2.3-tot-pslib030.UPF' in l
                                for l in lines)
            self.assertTrue(find_kind_ba2, "Unable to find the species line "
                                           "for Ba2")

            found1 = False
            found2 = False
            for l in lines:
                if 'starting_magnetization(1)' in l:
                    if found1:
                        raise ValueError(
                            "starting_magnetization(1) found multiple times")
                    found1 = True
                    self.assertAlmostEquals(
                        float(l.split('=')[1].replace('d', 'e')), 0.5)
                if 'starting_magnetization(2)' in l:
                    if found2:
                        raise ValueError(
                            "starting_magnetization(2) found multiple times")
                    found2 = True
                    self.assertAlmostEquals(
                        float(l.split('=')[1].replace('d', 'e')), -0.5)











# -*- coding: utf-8 -*-

from aiida.backends.utils import load_dbenv, is_dbenv_loaded

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

if not is_dbenv_loaded():
    load_dbenv()

from unittest import TestCase
from collections import Mapping
from aiida.workflows2.process import Process, ProcessSpec


class TestProcessSpec(TestCase):
    def setUp(self):
        self.spec = Process.spec()

    def test_get_inputs_template(self):
        s = ProcessSpec()
        s.input('a')
        s.input('b', default=5)

        template = s.get_inputs_template()
        self.assertIsInstance(template, Mapping)
        self._test_template(template)
        for attr in ['b']:
            self.assertTrue(
                attr in template,
                "Attribute '{}' not found in template".format(attr))

    def test_dynamic_input(self):
        from aiida.orm import Node
        from aiida.orm.data import Data

        n = Node()
        d = Data()
        port = self.spec.get_dynamic_input()
        self.assertFalse(port.validate("foo")[0])
        self.assertFalse(port.validate(5)[0])
        self.assertFalse(port.validate(n)[0])
        self.assertTrue(port.validate(d)[0])

    def test_dynamic_output(self):
        from aiida.orm import Node
        from aiida.orm.data import Data

        n = Node()
        d = Data()
        port = self.spec.get_dynamic_output()
        self.assertFalse(port.validate("foo")[0])
        self.assertFalse(port.validate(5)[0])
        self.assertFalse(port.validate(n)[0])
        self.assertTrue(port.validate(d)[0])

    def _test_template(self, template):
        template.a = 2
        self.assertEqual(template.a, 2)
        # Check the default is what we expect
        self.assertEqual(template.b, 5)
        with self.assertRaises(AttributeError):
            template.c = 6

        # Check that we can unpack
        self.assertEqual(dict(**template)['a'], 2)




# -*- coding: utf-8 -*-

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

def wf_check_input(*tuples):
    for arg, type in tuples:
        if not isinstance(arg,type):
            raise Exception("Argument %s is not of type %s" % (arg,type))

    return


@threadit
@workit_factory(code_type=Combine)
def create_PP_set(_calc, code_name=None, **db_PPs):
    """
    This function takes as input a dict of PseudoPot objects, and creates
    a PPSubset object which contains the PseudoPots. Extra links are created
    so that the user can access to the PseudoPots directly from the set. (AG)
    :param: dict of PP DB objects, the code_name allows the user to specify
    an existing code in the database to be linked to this calc by the wrapper.
    """

    uuid_list = []
    chemsys = []

    for el, PP in db_PPs.iteritems():
        # keep track of the UUIDs of the input PPs
        uuid_list.append(str(PP.uuid))
        # check argument types
        wf_check_input((PP, PseudoPot))
        # sanity check to ensure we have a PP consistent with the
        # elements we want
        if el != PP.content['element']:
            raise Exception("Discrepancy between element and PP")
        chemsys.append(el)

    # ensure we have at most one PP for a given element
    if len(chemsys) != len(set(chemsys)):
        raise Exception("Multiple PPs found for the same element")

    name = '_'.join(sorted(chemsys)) + '.pp_set'
    content = {'species': ','.join(sorted(chemsys)), 'PP_uuids': ','.join(sorted(uuid_list))}

    # create new PPSubset; since this is a work function, the existence
    # of a similar PPSubset will be taken care of by the wrapper itself
    # no need to use get or create here
    PP_set = PPSubset(name=name, content=content)

    # create extra links so that we can access the PPs directly from
    # the PP set without having to go through the CombineCalc.
    # These links are not set automatically by the wrapper, hence they
    # must be explicitly set by the user
    PP_set.roots.extend(db_PPs.values())

    return {'ppsubset': PP_set}


@threadit
@workit_factory(code_type=QEplugin)
def import_structures_from_QE_run(_calc, code_name=None, **qeoutputs):
    """
    This function upload Struc objects to the database,
    from a dict of QEOutput Data instances
    :param: a dict of QEOutput Data instances
    :return: a dict of the Struc objects added, {struc_name: Struc DB object}
    """

    cif_list = []

    for key, value in qeoutputs.iteritems():
        # check we are using output of a QE run as input
        wf_check_input((value, QEOutput))

        cif_name = str(value.name.replace('out', 'cif'))
        at = Atoms(positions=value.content['pos_fin'], cell=value.content['cell_fin'])
        at.set_chemical_symbols([str(el) for el in value.content['species']])

        # write to CIF file
        io.write(cif_name, at, format='cif')

        cif_list.append(cif_name)

    struc_list = import_structures_to_db(*cif_list)

    return struc_list


@threadit
@workit_factory(code_type=Manipulator)
def manipulate_struc(_calc, code_name=None, struc=None, transformation=None):
    """
    This function allows to generate a new structure by manipulating
    an already existing one. Manipulation includes randomizing atomic
    positions, apply a strain tensor, build supercells and so on.
    :param: a Struc to manipulate to generate a new one
    :return: dict with newly generated Struc
    """

    wf_check_input((struc, Struc), (transformation, ManipulateStruc))

    cif_name = str(struc.name.replace('out', ''))

    at = Atoms(positions=struc.content['positions'], cell=struc.content['cell'])
    at.set_chemical_symbols([str(el) for el in struc.content['species']])

    supercell_size = transformation.content.get('supercell',[1,1,1])

    at.repeat(supercell_size)

    cif_name += '_supercell_%s' % '_'.join([str(x) for x in supercell_size])

    if 'rattle' in transformation.content:
        stdev = transformation.content['rattle'][0]
        seed = transformation.content['rattle'][1]
        at.rattle(stdev=stdev, seed=seed)
        cif_name += '_rattled_%.5f' % stdev

    strain_tensor = transformation.content.get('strain', [[0.0,0.0,0.0],[0.0,0.0,0.0],[0.0,0.0,0.0]])

    at.set_cell(np.dot(np.array(strain_tensor)+np.identity(3), at.cell), scale_atoms=True)

    if np.array_equal(np.array(strain_tensor), np.zeros(3)) == False:
        cif_name += '_strained_%s' % str(strain_tensor)

    cif_name += '.cif'

    io.write(cif_name, at, format='cif')

    struc_list = import_structures_to_db(cif_name)

    return struc_list


@threadit
@workit_factory(code_type=QEplugin)
def create_QEInput_file(_calc, code_name=None, struc=None, qeinp=None, ppsubset=None):
    """
    This function takes as input a dict of DB objects (one Struc, one QEInput, one PP set)
    and creates a QE Input file object which contains the full qe file.
    :param: dict with Struc obejct, PPSubset object, QEInput object; ,
    the code_name allows the user to specify an existing code in the
    database to be linked to this calc by the wrapper.
    :return: dict with the FileData instance created
    """

    wf_check_input((struc, Struc), (qeinp, QEInput), (ppsubset, PPSubset))

    #...cut a lot of code here...

    content = {}
    content['filepath'] = work_dir + "/%s.in" % title
    content['filecontent'] = ss
    content['uuids'] =  ','.join(sorted(uuid_list))

    # create input file for QE calculation
    # if running locally, create the file
    # in the path specified
    if run_on_hal == False:
        fn = open(work_dir + "/%s.in" % title, 'w')
        fn.write(ss)
        fn.close()
    # if running remotely, we need to copy the file to the cluster
    elif run_on_hal == True:
        # create the file locally first in the working directory
        fn = open(os.getcwd()+"/%s.in" % title, 'w')
        fn.write(ss)
        fn.close()
        # create directory on cluster if it doesn't exist yet
        print "Creating directory on cluster...."
        # username and hostname taken from config.json
        create_dir(hal_username, hal_host, 'remote', work_dir)
        print "Done"
        # copy file to cluster
        print "Copying QE input file to cluster...."
        copy_file(hal_username, hal_host, 'local', os.getcwd(), work_dir, "%s.in" % title)
        print "Done"

    fileinp = FileData(name=title+'_input', content=content)

    return {'fileinp': fileinp}


@threadit
@workit_factory(code_type=QEplugin)
def run_QE_calc(_calc, code_name=None, fileinp=None):
    """
    This function takes as input a QE input file DB object: it runs the calculation
    (locally or remotely), then returns a FileData object containing information
    about the output file generated by the calculation. The parsing of the file
    will be done by a separate routine in order to allow for different parsers to
    be used.
    :param: dict with FileData object; the code_name allows the user to specify
    an existing code in the database to be linked to this calc by the wrapper.
    :return: dict with the FileData instance created
    """

    wf_check_input((fileinp, FileData))

    uuid_list = [str(fileinp.uuid)]

    # full path of the raw input file (local or remote)
    filepath = fileinp.content['filepath']
    # extract the filename from full path
    filename = filepath.split('/')[-1]
    # extract seed and working directory from filepath
    seed = filepath.replace('.in', '')
    cwd = '/'.join(seed.split('/')[:-1])

    # ...snip... code

    # create saga job
    container, js = run_job_saga(hal_username,'', 'local', executable,
            inputs = [filepath], name = None, project = None, queue = None,
            wall_time_limit = None, total_cpu_count = None, spmd_variation = None,
            workdir = [cwd], outputs = [seed+'.out'], errors = [seed+'.err'])

    # run the saga job
    container.run()

    # get jobs id
    job_id_list = [job.id for job in container.jobs]

    # store jobs id into content of the calc object in database
    # this is important to leave a handle
    s = Session()
    self_copy = s.merge(self)
    # in case the content field is empty
    if not self_copy.content:
        self_copy.content = {}
    self_copy.content['jobs_id'] = job_id_list
    flag_modified(self_copy, "content")
    s.commit()

    # test: periodically check number of jobs running
    while any([x in running_states for x in container.get_states()]):
        running_jobs = [y for y in container.jobs if y.get_state() in running_states]
        print "Number of running jobs: %d" % len(running_jobs)

        time.sleep(11)

    res_list = []

    # when jobs are finished, get states of the jobs
    # this is important in order to determine if the
    # output will be in a Final or Partial state
    for state in container.get_states():
        if state == saga.job.DONE:
            res_list.append("Finished")
        elif state == saga.job.FAILED:
            res_list.append("Failed")
        elif state == saga.job.CANCELED:
            res_list.append("Canceled")

    print res_list

    content = {}
    content['filepath'] = seed+'.out'
    # currently assuming one job per saga task
    # convention: job finished -> data is final
    # otherwise data is partial
    # validity will be checked automatically by methods invoked by wrapper
    if res_list[0] == "Finished":
        state = "Final"
    else:
        state = "Partial"

    title = seed.split('/')[-1]
    # save FileData object to database; no need to check for duplicates since
    # the wrapper will take care of fast-forwarding if calculation with the same
    # inputs is found in the database
    fileout = FileData(name=title+'_output', state=state, content=content)

    # if calculation is performed on the cluster, copy the file back to local directory
    if run_on_hal == True:
        print "Copying output file from cluster...."
        copy_file(hal_username, hal_host, 'remote', os.getcwd(), cwd, filename.replace('.in', '.out'))
        print "Done"

    return {title: fileout}


@threadit
@workit_factory(code_type=QEparser)
def parsing_QE_output(_calc, code_name=None, struc=None, ppsubset=None,
    parser=None, observed=None):
    """
    This function takes as input a FileData object with the location of the output
    file generated by QE. It then parses the file and store the relevant info into a
    Data object. The use of the code_name allows to specify the parser to be used.
    Currently only parsing using espresso python module is allowed. (AG)
    :param: FileData object created by QE calc
    :return: dict with QEOutput object containing information about parsed QE output.
    TO-DO: implement different parsing options
    """

    wf_check_input((struc, Struc), (ppsubset, PPSubset), (parser, ParserParam),
        (observed, Observed))

    uuid_list = [str(struc.uuid), str(ppsubset.uuid), str(parser.uuid), str(observed.uuid)]

    # currently this is the only parsing option supported
    if parser.name != 'Espresso wrapper by Zhongnan Xu':
        raise Exception("{} not currently supported".format(parser))
    else:
        print "Analyzing QE output with {}".format(parser)

    title = observed.content['filepath'].split('/')[-1]

    if run_on_hal == False:
        seed = observed.content['filepath'].replace('.out','')
    # if calculation is run on cluster, the analysis is performed
    # on the local files that are stored in the current working
    # directory: the output file is copied from cluster at the end
    # of the QE calculation, before parsing takes place
    elif run_on_hal == True:
        seed = os.getcwd() + '/' + title

    content = {}
    dict_PPs = {}

    try:
        # using espresso module with some changes done by me
        # in order to initialise Espresso instance from filename
        e = Espresso(filename=seed)
        # True if calculation is finished
        content['calc_finished'] = e.calc_finished
        # True if calculation convergenced
        content['converged'] = e.converged
        # check for electronic convergence
        content['electronic_converged'] = e.electronic_converged
        content['total_energy'] = e.energy_free
        content['energy_atom'] = e.energy_free/e.int_params['nat']

        # ...snip....
    except:
        # default if file not found and/or other parsing problems
        content['calc_finished'] = False

    res = QEOutput(name=title, content=content)

    return {title: res}


@threadit
@workit_factory(code_type=QEplugin)
def observeQE(_calc, code_name=None, fileinp=None):
    """
    This function is a prototype for an observer calc for QE. It periodically checks
    the calculation output and stops it if something goes wrong according to
    the logic set by the user.
    :param: dict with FileData object; the code_name allows the user to specify
    an existing code in the database to be linked to this calc by the wrapper.
    :return: dict with the Observed instance created
    """

    wf_check_input((fileinp, FileData))

    content = {}

    uuid_list = [str(fileinp.uuid)]

    filepath = fileinp.content['filepath']
    filecontent = fileinp.content['filecontent']

    out_file = filepath.replace('.in', '.out')

    number_of_tries = 0
    # the observer runs the QE calculation, and then periodically looks at it
    # try to submit the saga job a max number of times
    # different threads submitting saga jobs through the same channel
    # may lead to connection errors raised by saga
    while number_of_tries <= max_number_retries:
        try:
            res = run_QE_calc(_calc, code_name='Quantum Espresso 5.1.2 Mac OS X 10.9',
                fileinp=fileinp)
        except:
            number_of_tries += 1
            time.sleep(10)
            continue
        break

    if number_of_tries == max_number_retries:
        raise Exception("Saga job submission failed after %d attempts" % max_number_retries)

    # in threading, res is now a Thread object because we haven't blocked
    # with w4 yet, so we can access the output before the function actually
    # finishes the execution
    # ensures we don't check anymore if the thread finishes without problems
    for key, value in res.iteritems():
        if isinstance(value, threading.Thread) or isinstance(value, multiprocessing.Process):
            thread = value
        else:
            thread = None

    if thread is not None:
        while thread.is_alive():
            # necessary because we need first to wait for the output
            # file to be created, before examining it
            if os.path.isfile(out_file):
                fn = open(out_file, 'r')
                hh = fn.readlines()
                fn.close()
                # dummy prototype for checking: print last 10 lines
                # in production, call the function that does the required
                # checks on the file
                print "Last 10 output lines:\n%s" % "".join(hh[-10:])
                # assume one slave per observer
                slave = _calc.slaves[0]
                slave.kill_itself('gra2pal', '', 'local')
            else:
                print "Waiting for output file to be created....."
            time.sleep(30)

    # get the result only when the thread has finished/has been killed
    final_res = w4(res)

    content['uuids'] =  ','.join(sorted(uuid_list))
    content['filepath'] = out_file

    # TO-DO: maybe change state to Partial according to what happened to
    # the thread. We may have a flag to change the state accordingly
    observer_res = Observed(name='observed', state='Final', content=content)

    return {'observed': observer_res}


@threadit
@workit_factory(code_type=QEplugin)
def QE_workflow(_calc, code_name=None, struc=None,
    qeinp=None, ppsubset=None, parser=None):
    """
    Example of workflow to run an energy calculation in Quantum Espresso. (AG)
    :param: Struc, QEInput, PPSubset, Parser.
    :return: dict with QEOutput, i.e. parsed QE output file
    """

    wf_check_input((struc, Struc), (qeinp, QEInput), (ppsubset, PPSubset), (parser, ParserParam))

    res = w4(create_QEInput_file(_calc, struc=struc, qeinp=qeinp, ppsubset=ppsubset))

    # not observed version
    #out = w4(run_QE_calc(self, code_name='Quantum Espresso 5.1.2 Mac OS X 10.9',
    #            fileinp=fileinp))

    # version with observer
    qe_out = w4(observeQE(_calc, **res))

    observed = qe_out['observed']

    qe_out_parsed = w4(parsing_QE_output(_calc, 'Espresso wrapper by Zhongnan Xu',
        struc=struc, ppsubset=ppsubset, parser=parser, observed=observed))

    # setting up direct links for easier queries
    qe_out_parsed.values()[0].roots.extend([struc, qeinp, ppsubset, parser])

    return qe_out_parsed


# for testing the workflow
if __name__ == "__main__":

    s = Session()
    # retrieve entries that are already in the DB
    # we can have functions that generate this entries automatically
    struc = s.query(Struc).filter(Struc.content['formula'].cast(String) == 'Li3').\
        filter(Struc.name=='Li_C19_coarse_50Ry').first()
    qeinp = s.query(QEInput).filter(QEInput.validity != 'Bad').filter(QEInput.content['control','title'].\
        cast(String) == 'Li_C19_coarse_50Ry_scf').filter(QEInput.content['ions']!=None).first()
    pp = s.query(PseudoPot).filter(PseudoPot.content['element'].cast(String) == 'Li').first()
    parser = s.query(ParserParam).filter(ParserParam.name == 'Espresso wrapper by Zhongnan Xu').first()

    dict_pp = {'Li': pp}

    caller1 = Calc(name='PP set generator')

    res = w4(create_PP_set(caller1, **dict_pp))

    ppsubset = res['ppsubset']

    caller2 = Calc(name='QE_workflow')

    # call the workflow
    res = w4(QE_workflow(caller2, struc=struc, qeinp=qeinp, ppsubset=ppsubset, parser=parser))
    #res = w4(equation_of_state_workflow(caller2, struc=struc, ppsubset=ppsubset, parser=parser))

    s.close()

    # explicitly close the job service (should be closed
    # automatically by saga anyway)
    close_js(hal_username, '', 'local')


# -*- coding: utf-8 -*-
from aiida.common.extendeddicts import *
import unittest
import copy

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class TestFFADExample(FixedFieldsAttributeDict):
    """
    An example class that accepts only the 'a', 'b' and 'c' keys/attributes.
    """
    _valid_fields = ('a', 'b', 'c')


class TestDFADExample(DefaultFieldsAttributeDict):
    """
    An example class that has 'a', 'b' and 'c' as default keys.
    """
    _default_fields = ('a', 'b', 'c')

    def validate_a(self, value):
        # Ok if unset
        if value is None:
            return

        if not isinstance(value, (int, long)):
            raise TypeError('expecting integer')
        if value < 0:
            raise ValueError('expecting a positive or zero value')


class TestAttributeDictAccess(unittest.TestCase):
    """
    Try to access the dictionary elements in various ways, copying (shallow and
    deep), check raised exceptions.
    """

    def test_access_dict_to_attr(self):
        d = AttributeDict()
        d['test'] = 'abc'
        self.assertEquals(d.test, 'abc')

    def test_access_attr_to_dict(self):
        d = AttributeDict()
        d.test = 'def'
        self.assertEquals(d['test'], 'def')

    def test_access_nonexisting_asattr(self):
        d = AttributeDict()
        with self.assertRaises(AttributeError):
            a = d.test

    def test_access_nonexisting_askey(self):
        d = AttributeDict()
        with self.assertRaises(KeyError):
            a = d['test']

    def test_del_nonexisting_askey(self):
        d = AttributeDict()
        with self.assertRaises(KeyError):
            del d['test']

    def test_del_nonexisting_asattr(self):
        d = AttributeDict()
        with self.assertRaises(AttributeError):
            del d.test

    def test_copy(self):
        d1 = AttributeDict()
        d1.x = 'a'
        d2 = d1.copy()
        d2.x = 'b'
        self.assertEquals(d1.x, 'a')
        self.assertEquals(d2.x, 'b')

    def test_delete_after_copy(self):
        d1 = AttributeDict()
        d1.x = 'a'
        d1.y = 'b'
        d2 = d1.copy()
        del d1.x
        del d1['y']
        with self.assertRaises(AttributeError):
            _ = d1.x
        with self.assertRaises(KeyError):
            _ = d1['y']
        self.assertEquals(d2['x'], 'a')
        self.assertEquals(d2.y, 'b')
        self.assertEquals(set(d1.keys()), set({}))
        self.assertEquals(set(d2.keys()), set({'x', 'y'}))

    def test_shallowcopy1(self):
        d1 = AttributeDict()
        d1.x = [1, 2, 3]
        d1.y = 3
        d2 = d1.copy()
        d2.x[0] = 4
        d2.y = 5
        self.assertEquals(d1.x, [4, 2, 3])  # copy does a shallow copy
        self.assertEquals(d2.x, [4, 2, 3])
        self.assertEquals(d1.y, 3)
        self.assertEquals(d2.y, 5)

    def test_shallowcopy2(self):
        """
        Also test access at nested levels
        """
        d1 = AttributeDict()
        d1.x = {'a': 'b', 'c': 'd'}
        # d2 = copy.deepcopy(d1)
        d2 = d1.copy()
        # doesn't work like this, would work as d2['x']['a']
        # i think that it is because deepcopy on dict actually creates a
        # copy only if the data is changed; but for a nested dict,
        # d2.x returns a dict wrapped in our class and this looses all the
        # information on what should be updated when changed.
        d2.x['a'] = 'ggg'
        self.assertEquals(d1.x['a'], 'ggg')  # copy does a shallow copy
        self.assertEquals(d2.x['a'], 'ggg')

    def test_deepcopy1(self):
        d1 = AttributeDict()
        d1.x = [1, 2, 3]
        d2 = copy.deepcopy(d1)
        d2.x[0] = 4
        self.assertEquals(d1.x, [1, 2, 3])
        self.assertEquals(d2.x, [4, 2, 3])

    def test_shallowcopy2(self):
        """
        Also test access at nested levels
        """
        d1 = AttributeDict()
        d1.x = {'a': 'b', 'c': 'd'}
        d2 = copy.deepcopy(d1)
        d2.x['a'] = 'ggg'
        self.assertEquals(d1.x['a'], 'b')  # copy does a shallow copy
        self.assertEquals(d2.x['a'], 'ggg')


class TestAttributeDictNested(unittest.TestCase):
    """
    Test the functionality of nested AttributeDict classes.
    """

    def test_nested(self):
        d1 = AttributeDict({'x': 1, 'y': 2})
        d2 = AttributeDict({'z': 3, 'w': 4})
        d1.nested = d2
        self.assertEquals(d1.nested.z, 3)
        self.assertEquals(d1['nested'].w, 4)
        self.assertEquals(d1.nested['w'], 4)
        d2['w'] = 5
        self.assertEquals(d1['nested'].w, 5)
        self.assertEquals(d1.nested['w'], 5)

    def test_comparison(self):
        d1 = AttributeDict({'x': 1, 'y': 2, 'z': AttributeDict({'w': 3})})
        d2 = AttributeDict({'x': 1, 'y': 2, 'z': AttributeDict({'w': 3})})

        # They compare to the same value but they are different objects
        self.assertFalse(d1 is d2)
        self.assertEquals(d1, d2)

        d2.z.w = 4
        self.assertNotEquals(d1, d2)

    def test_nested_deepcopy(self):
        d1 = AttributeDict({'x': 1, 'y': 2})
        d2 = AttributeDict({'z': 3, 'w': 4})
        d1.nested = d2
        d1copy = copy.deepcopy(d1)
        self.assertEquals(d1copy.nested.z, 3)
        self.assertEquals(d1copy['nested'].w, 4)
        self.assertEquals(d1copy.nested['w'], 4)
        d2['w'] = 5
        self.assertEquals(d1copy['nested'].w, 4)  # Nothing has changed
        self.assertEquals(d1copy.nested['w'], 4)  # Nothing has changed
        self.assertEquals(d1copy.nested.w, 4)  # Nothing has changed

        self.assertEquals(d1['nested'].w, 5)  # The old one is updated
        self.assertEquals(d1.nested['w'], 5)  # The old one is updated
        self.assertEquals(d1.nested.w, 5)  # The old one is updated

        d1copy.nested.w = 6
        self.assertEquals(d1copy.nested.w, 6)
        self.assertEquals(d1.nested.w, 5)
        self.assertEquals(d2.w, 5)


class TestAttributeDictSerialize(unittest.TestCase):
    """
    Test serialization/deserialization (with json, pickle, ...)
    """

    def test_json(self):
        import json

        d1 = AttributeDict({'x': 1, 'y': 2})
        d2 = json.loads(json.dumps(d1))
        # Note that here I am comparing a dictionary (d2) with a
        # AttributeDict (d2) and they still compare to equal
        self.assertEquals(d1, d2)

    def test_json_recursive(self):
        import json

        d1 = AttributeDict({'x': 1, 'y': 2, 'z': AttributeDict({'w': 4})})
        d2 = json.loads(json.dumps(d1))
        # Note that here I am comparing a dictionary (d2) with a (recursive)
        # AttributeDict (d2) and they still compare to equal
        self.assertEquals(d1, d2)

    def test_pickle(self):
        import pickle

        d1 = AttributeDict({'x': 1, 'y': 2})
        d2 = pickle.loads(pickle.dumps(d1))
        self.assertEquals(d1, d2)

    def test_pickle_recursive(self):
        import pickle

        d1 = AttributeDict({'x': 1, 'y': 2, 'z': AttributeDict({'w': 4})})
        d2 = pickle.loads(pickle.dumps(d1))
        self.assertEquals(d1, d2)


class TestFFAD(unittest.TestCase):
    def test_insertion(self):
        a = TestFFADExample()
        a['a'] = 1
        a.b = 2
        # Not a valid key.
        with self.assertRaises(KeyError):
            a['d'] = 2
        with self.assertRaises(AttributeError):
            a.e = 5

    def test_insertion_on_init(self):
        a = TestFFADExample({'a': 1, 'b': 2})
        with self.assertRaises(KeyError):
            # 'd' is not a valid key
            a = TestFFADExample({'a': 1, 'd': 2})

    def test_pickle(self):
        """
        Note: pickle works here because self._valid_fields is defined
        at the class level!
        """
        import pickle

        a = TestFFADExample({'a': 1, 'b': 2})
        b = pickle.loads(pickle.dumps(a))
        b.c = 3
        with self.assertRaises(KeyError):
            b['d'] = 2

    def test_class_attribute(self):
        """
        I test that the get_valid_fields() is working as a class method,
        so I don't need to instantiate the class to get the list.
        """
        self.assertEquals(set(TestFFADExample.get_valid_fields()),
                          set(['a', 'b', 'c']))


class TestDFAD(unittest.TestCase):
    def test_insertion_and_retrieval(self):
        a = TestDFADExample()
        a['a'] = 1
        a.b = 2
        a['d'] = 3
        a.e = 4
        self.assertEquals(a.a, 1)
        self.assertEquals(a.b, 2)
        self.assertEquals(a['d'], 3)
        self.assertEquals(a['e'], 4)

    def test_keylist_method(self):
        a = TestDFADExample()
        a['a'] = 1
        a.b = 2
        a['d'] = 3
        a.e = 4

        self.assertEquals(set(a.get_default_fields()), set(['a', 'b', 'c']))
        self.assertEquals(set(a.keys()), set(['a', 'b', 'd', 'e']))
        self.assertEquals(set(a.defaultkeys()), set(['a', 'b']))
        self.assertEquals(set(a.extrakeys()), set(['d', 'e']))
        self.assertIsNone(a.c)

    def test_class_attribute(self):
        """
        I test that the get_default_fields() is working as a class method,
        so I don't need to instantiate the class to get the list.
        """
        self.assertEquals(set(TestDFADExample.get_default_fields()),
                          set(['a', 'b', 'c']))


    def test_validation(self):
        o = TestDFADExample()

        # Should be ok to have an empty 'a' attribute
        o.validate()

        o.a = 4
        o.b = 'text'

        # This should be fine
        o.validate()

        o.a = 'string'
        # o.a must be a positive integer
        with self.assertRaises(ValidationError):
            o.validate()

        o.a = -3
        # a.a must be a positive integer
        with self.assertRaises(ValidationError):
            o.validate()




# -*- coding: utf-8 -*-

from aiida.orm.calculation.job.codtools.ciffilter import CiffilterCalculation

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class CifsplitprimitiveCalculation(CiffilterCalculation):
    """
    Specific input plugin for cif_split_primitive from cod-tools package.
    """

    def _init_internal_params(self):
        super(CifsplitprimitiveCalculation, self)._init_internal_params()

        self._default_parser = 'codtools.cifsplitprimitive'

        self._SPLIT_DIR = 'split'

    def _prepare_for_submission(self, tempfolder, inputdict):
        import os

        calcinfo = super(CifsplitprimitiveCalculation,
                         self)._prepare_for_submission(tempfolder, inputdict)

        split_dir = tempfolder.get_abs_path(self._SPLIT_DIR)
        os.mkdir(split_dir)

        calcinfo.codes_info[0].cmdline_params.extend(['--output-dir', self._SPLIT_DIR])
        calcinfo.retrieve_list.append(self._SPLIT_DIR)
        
        return calcinfo



# -*- coding: utf-8 -*-
from aiida.backends.djsite.db.testbase import AiidaTestCase

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class CalcStateChanges(AiidaTestCase):
    # Class to check if the migration code that deals with removing the
    # NOTFOUND and UNDETERMINED calc states works properly
    def test_unexpected_calc_states(self):
        import logging

        from django.utils import timezone
        from aiida.orm.calculation import Calculation

        # Have to use this ugly way of importing because the django migration
        # files start with numbers which are not a valid package name
        state_change = __import__(
            'aiida.backends.djsite.db.migrations.0002_db_state_change',
            fromlist=['fix_calc_states']
        )
        from aiida.common.datastructures import calc_states
        from aiida.backends.djsite.db.models import DbCalcState, DbLog
        from aiida.orm.calculation.job import JobCalculation

        calc_params = {
            'computer': self.computer,
            'resources': {'num_machines': 1,
                          'num_mpiprocs_per_machine': 1}
        }

        for state in ['NOTFOUND', 'UNDETERMINED']:
            # Let's create a dummy job calculation
            job = JobCalculation(**calc_params)
            job.store()
            # Now save the errant state
            DbCalcState(dbnode=job.dbnode, state=state).save()

            time_before_fix = timezone.now()

            # Temporarily disable logging to the stream handler (i.e. screen)
            # because otherwise fix_calc_states will print warnings
            handler = next((h for h in logging.getLogger('aiida').handlers if
                           isinstance(h, logging.StreamHandler)), None)
            if handler:
                handler.setLevel(logging.ERROR)

            # Call the code that deals with updating these states
            state_change.fix_calc_states(None, None)

            if handler:
                handler.setLevel(logging.NOTSET)

            current_state = job.get_state()
            self.assertNotEqual(current_state, state,
                                "Migration code failed to change removed state {}".
                                format(state))
            self.assertEqual(current_state, calc_states.FAILED,
                             "Migration code failed to set removed state {} to {}".
                             format(current_state, calc_states.FAILED))

            result = DbLog.objects.filter(
                objpk__exact=job.pk,
                levelname__exact=logging.getLevelName(logging.WARNING),
                time__gt=time_before_fix
            )
            self.assertEquals(len(result), 1,
                              "Couldn't find a warning message with the change"
                              "from {} to {}".format(state, calc_states.FAILED)
                              )



# -*- coding: utf-8 -*-
"""
An IPython extension that provides a magic command to load
basic aiida commands.

This makes it much easier to start.

Produces output in:

* Plaintext (IPython [qt]console)
* HTML (IPython notebook, ``nbconvert --to html``, ``--to slides``)
* JSON (IPython notebook ``.ipynb`` files)
* LaTeX (e.g. ``ipython nbconvert example.ipynb --to LaTeX --post PDF``)


Notes on how to load it at start:
https://ipython.org/ipython-doc/3/config/intro.html

Usage
======

.. sourcecode:: ipython

   In [1]: %load_ext aiida_magic

   In [2]: %aiida
"""
import json
import IPython
from IPython.core.magic import (magics_class, line_magic, Magics, 
                                needs_local_scope)

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

def add_to_ns(local_ns, name, obj):
    """
    Add a new variable with name ``name`` and value ``obj`` to the 
    namespace ``local_ns``, optionally showing a warning if we are 
    hiding an existing variable.
    
    .. todo:: implement the warning.
    
    Example::

        # assuming that local_ns is a dictionary, e.g. from locals()
        import sys
        add_to_ns(local_ns, 'sys', sys)
    """
    if name in local_ns:
        # TODO: print warning, or raise
        pass
    local_ns[name] = obj
    
@magics_class
class AiiDALoaderMagics(Magics):

    @needs_local_scope
    @line_magic
    def aiida(self, line='', local_ns=None):
        """Load AiiDA in ipython (checking if it was already loaded), and
        inserts in the namespace the main AiiDA classes (the same that are
        loaded in ``verdi shell``.

        Usage::

            %aiida [optional parameters]
        
        .. todo:: implement parameters, e.g. for the profile to load.
        """
        import aiida
        
        from aiida import is_dbenv_loaded, load_dbenv
                
        self.is_warning = False
        if is_dbenv_loaded():
            self.current_state = "Note! AiiDA DB environment already loaded! I do not reload it again."
            self.is_warning = True
        else:
            load_dbenv()
            self.current_state = "Loaded AiiDA DB environment."

        from aiida.cmdline.commands.shell import Shell
        user_ns = Shell().get_start_namespace()
        for k, v in user_ns.iteritems():
            add_to_ns(local_ns, k, v)
            
        return self

    def _repr_json_(self):
        """
        Output in JSON format.
        """
        obj = {
            'current_state': self.current_state}
        if IPython.version_info[0] >= 3:
            return obj
        else:
            return json.dumps(obj)

    def _repr_html_(self):
        """
        Output in HTML format.
        """
        html = "<p>"
        if self.is_warning:
            html += "<strong>"
        html += self.current_state
        if self.is_warning:
            html += "</strong>"
        html += "</p>"

        return html

    def _repr_latex_(self):
        """
        Output in LaTeX format.
        """
        if self.is_warning:
            latex = "\\emph{%s}\n" % self.current_state
        else:
            latex = "%s\n" % self.current_state

        return latex

    def _repr_pretty_(self, pp, cycle):
        """
        Output in text format.
        """
        if self.is_warning:
            warning_str = "** "
        else:
            warning_str = ""
        text = "%s%s\n" % (warning_str, self.current_state)

        pp.text(text)

        
def load_ipython_extension(ipython):
    """
    Triggers the load of all the AiiDA magic commands.
    """
    ipython.register_magics(AiiDALoaderMagics)



# -*- coding: utf-8 -*-

from aiida.backends.utils import load_dbenv, is_dbenv_loaded

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

if not is_dbenv_loaded():
    load_dbenv()

from unittest import TestCase
from aiida.workflows2.process import Process
from aiida.workflows2.db_types import Int
from workflows2.common import ProcessScope, DummyProcess, BadOutput
from aiida.common.lang import override
import uuid
import threading


class ProcessStackTest(Process):
    @override
    def _run(self):
        pass

    @override
    def on_create(self, pid, inputs=None):
        super(ProcessStackTest, self).on_create(pid, inputs)
        self._thread_id = threading.current_thread().ident

    @override
    def on_recreate(self, pid, saved_instance_state):
        super(ProcessStackTest, self).on_recreate(pid, saved_instance_state)
        self._thread_id = threading.current_thread().ident

    @override
    def on_stop(self):
        # The therad must match the one used in on_create because process
        # stack is using thread local storage to keep track of who called who
        super(ProcessStackTest, self).on_stop()
        assert self._thread_id is threading.current_thread().ident


class TestProcess(TestCase):
    def test_process_stack(self):
        ProcessStackTest.run()

    def test_inputs(self):
        with self.assertRaises(AssertionError):
            BadOutput.run()

    def test_pid_uuid(self):
        with ProcessScope(DummyProcess(store_provenance=False)) as p:
            self.assertEqual(uuid.UUID(p._calc.uuid), p.pid)

    def test_input_link_creation(self):
        inputs = ["1", "2", "3", "4"]

        with ProcessScope(
                DummyProcess(store_provenance=True),
                inputs={l: Int(l) for l in inputs}) as p:

            for label, value in p._calc.get_inputs_dict().iteritems():
                self.assertTrue(label in inputs)
                self.assertEqual(int(label), int(value.value))
                inputs.remove(label)

            # Make sure there are no other inputs
            self.assertFalse(inputs)




# -*- coding: utf-8 -*-
from aiida.common.pluginloader import BaseFactory

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


def CalculationFactory(module, from_abstract=False):
    """
    Return a suitable JobCalculation subclass.

    :param module: a valid string recognized as a Calculation plugin
    :param from_abstract: A boolean. If False (default), actually look only
      to subclasses to JobCalculation, not to the base Calculation class.
      If True, check for valid strings for plugins of the Calculation base class.
    """
    from aiida.orm.calculation import Calculation
    from aiida.orm.calculation.job import JobCalculation

    if from_abstract:
        return BaseFactory(module, Calculation, "aiida.orm.calculation")
    else:
        return BaseFactory(module, JobCalculation, "aiida.orm.calculation.job",
                           suffix="Calculation")


def DataFactory(module):
    """
    Return a suitable Data subclass.
    """
    from aiida.orm.data import Data

    return BaseFactory(module, Data, "aiida.orm.data")


def WorkflowFactory(module):
    """
    Return a suitable Workflow subclass.
    """
    from aiida.orm.workflow import Workflow

    return BaseFactory(module, Workflow, "aiida.workflows")


def load_node(node_id=None, pk=None, uuid=None, parent_class=None):
    """
    Return an AiiDA node given PK or UUID.

    :param node_id: PK (integer) or UUID (string) or a node
    :param pk: PK of a node
    :param uuid: UUID of a node
    :param parent_class: if specified, checks whether the node loaded is a
        subclass of parent_class
    :return: an AiiDA node
    :raise ValueError: if none or more than one of parameters is supplied
        or type of node_id is neither string nor integer.
    :raise NotExistent: if the parent_class is specified
        and no matching Node is found.
    """
    from aiida.common.exceptions import NotExistent
    # This must be done inside here, because at import time the profile
    # must have been already loaded. If you put it at the module level,
    # the implementation is frozen to the default one at import time.
    from aiida.orm.implementation import Node


    if int(node_id is None) + int(pk is None) + int(uuid is None) == 3:
        raise ValueError("one of the parameters 'node_id', 'pk' and 'uuid' "
                         "has to be supplied")
    if int(node_id is None) + int(pk is None) + int(uuid is None) < 2:
        raise ValueError("only one of parameters 'node_id', 'pk' and 'uuid' "
                         "has to be supplied")
    loaded_node = None
    if node_id is not None:
        if isinstance(node_id, str) or isinstance(node_id, unicode):
            loaded_node = Node.get_subclass_from_uuid(node_id)
        elif isinstance(node_id, int):
            loaded_node = Node.get_subclass_from_pk(node_id)
        else:
            raise ValueError("'node_id' has to be either string, unicode or "
                             "integer, {} given".format(type(node_id)))
    if loaded_node is None:
        if pk is not None:
            loaded_node = Node.get_subclass_from_pk(pk)
        else:
            loaded_node = Node.get_subclass_from_uuid(uuid)

    if parent_class is not None:
        if not issubclass(parent_class, Node):
            raise ValueError("parent_class must be a subclass of Node")
        if not isinstance(loaded_node, parent_class):
            raise NotExistent('No node found as '
                              'subclass of {}'.format(parent_class))

    return loaded_node


def load_workflow(wf_id=None, pk=None, uuid=None):
    """
    Return an AiiDA workflow given PK or UUID.

    :param wf_id: PK (integer) or UUID (string) or a workflow
    :param pk: PK of a workflow
    :param uuid: UUID of a workflow
    :return: an AiiDA workflow
    :raises: ValueError if none or more than one of parameters is supplied
        or type of wf_id is neither string nor integer
    """
    # This must be done inside here, because at import time the profile
    # must have been already loaded. If you put it at the module level,
    # the implementation is frozen to the default one at import time.
    from aiida.orm.implementation import Workflow

    if int(wf_id is None) + int(pk is None) + int(uuid is None) == 3:
        raise ValueError("one of the parameters 'wf_id', 'pk' and 'uuid' "
                         "has to be supplied")
    if int(wf_id is None) + int(pk is None) + int(uuid is None) < 2:
        raise ValueError("only one of parameters 'wf_id', 'pk' and 'uuid' "
                         "has to be supplied")
    if wf_id is not None:
        if isinstance(wf_id, str) or isinstance(wf_id, unicode):
            return Workflow.get_subclass_from_uuid(wf_id)
        elif isinstance(wf_id, int):
            return Workflow.get_subclass_from_pk(wf_id)
        else:
            raise ValueError("'wf_id' has to be either string, unicode or "
                             "integer, {} given".format(type(wf_id)))
    if pk is not None:
        return Workflow.get_subclass_from_pk(pk)
    else:
        return Workflow.get_subclass_from_uuid(uuid)



# -*- coding: utf-8 -*-

import datetime
from django.db import transaction, IntegrityError
from django.db.models import Q
from aiida.common.utils import str_timedelta
from aiida.common.datastructures import sort_states, calc_states
from aiida.common.exceptions import ModificationNotAllowed, DbContentError
from aiida.backends.djsite.utils import get_automatic_user
from aiida.orm.group import Group
from aiida.orm.implementation.django.calculation import Calculation
from aiida.orm.implementation.general.calculation.job import (
    AbstractJobCalculation)
from aiida.common.pluginloader import from_type_to_pluginclassname
from aiida.utils import timezone

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"


class JobCalculation(AbstractJobCalculation, Calculation):
    def _set_state(self, state):
        """
        Set the state of the calculation.

        Set it in the DbCalcState to have also the uniqueness check.
        Moreover (except for the IMPORTED state) also store in the 'state'
        attribute, useful to know it also after importing, and for faster
        querying.

        .. todo:: Add further checks to enforce that the states are set
           in order?

        :param state: a string with the state. This must be a valid string,
          from ``aiida.common.datastructures.calc_states``.
        :raise: ModificationNotAllowed if the given state was already set.
        """

        from aiida.common.datastructures import sort_states
        from aiida.backends.djsite.db.models import DbCalcState

        if not self.is_stored:
            raise ModificationNotAllowed("Cannot set the calculation state "
                                         "before storing")

        if state not in calc_states:
            raise ValueError(
                "'{}' is not a valid calculation status".format(state))

        old_state = self.get_state()
        if old_state:
            state_sequence = [state, old_state]

            # sort from new to old: if they are equal, then it is a valid
            # advance in state (otherwise, we are going backwards...)
            if sort_states(state_sequence) != state_sequence:
                raise ModificationNotAllowed("Cannot change the state from {} "
                                             "to {}".format(old_state, state))

        try:
            with transaction.commit_on_success():
                new_state = DbCalcState(dbnode=self.dbnode, state=state).save()
        except IntegrityError:
            raise ModificationNotAllowed(
                "Calculation pk= {} already transited through "
                "the state {}".format(self.pk, state))

        # For non-imported states, also set in the attribute (so that, if we
        # export, we can still see the original state the calculation had.
        if state != calc_states.IMPORTED:
            self._set_attr('state', state)

    def get_state(self, from_attribute=False):
        """
        Get the state of the calculation.

        .. note:: this method returns the NOTFOUND state if no state
          is found in the DB.

        .. note:: the 'most recent' state is obtained using the logic in the
          ``aiida.common.datastructures.sort_states`` function.

        .. todo:: Understand if the state returned when no state entry is found
          in the DB is the best choice.

        :param from_attribute: if set to True, read it from the attributes
          (the attribute is also set with set_state, unless the state is set
          to IMPORTED; in this way we can also see the state before storing).

        :return: a string. If from_attribute is True and no attribute is found,
          return None. If from_attribute is False and no entry is found in the
          DB, return the "NOTFOUND" state.
        """
        from aiida.backends.djsite.db.models import DbCalcState
        if from_attribute:
            return self.get_attr('state', None)
        else:
            if not self.is_stored:
                return calc_states.NEW
            else:
                this_calc_states = DbCalcState.objects.filter(
                    dbnode=self).values_list('state', flat=True)
                if not this_calc_states:
                    return None
                else:
                    try:
                        most_recent_state = sort_states(this_calc_states)[0]
                    except ValueError as e:
                        raise DbContentError("Error in the content of the "
                                             "DbCalcState table ({})".format(
                            e.message))

                    return most_recent_state

    @classmethod
    def _list_calculations_old(cls, states=None, past_days=None, group=None,
                               group_pk=None, all_users=False, pks=[],
                               relative_ctime=True):
        """
        Return a string with a description of the AiiDA calculations.

        .. todo:: does not support the query for the IMPORTED state (since it
          checks the state in the Attributes, not in the DbCalcState table).
          Decide which is the correct logi and implement the correct query.

        :param states: a list of string with states. If set, print only the
            calculations in the states "states", otherwise shows all.
            Default = None.
        :param past_days: If specified, show only calculations that were
            created in the given number of past days.
        :param group: If specified, show only calculations belonging to a
            user-defined group with the given name.
            Can use colons to separate the group name from the type,
            as specified in :py:meth:`aiida.orm.group.Group.get_from_string`
            method.
        :param group_pk: If specified, show only calculations belonging to a
            user-defined group with the given PK.
        :param pks: if specified, must be a list of integers, and only
            calculations within that list are shown. Otherwise, all
            calculations are shown.
            If specified, sets state to None and ignores the
            value of the ``past_days`` option.")
        :param relative_ctime: if true, prints the creation time relative from now.
                               (like 2days ago). Default = True
        :param all_users: if True, list calculation belonging to all users.
                           Default = False

        :return: a string with description of calculations.
        """
        # I assume that calc_states are strings. If this changes in the future,
        # update the filter below from dbattributes__tval to the correct field.
        from aiida.backends.djsite.db.models import DbAuthInfo, DbAttribute
        from aiida.daemon.timestamps import get_last_daemon_timestamp

        if states:
            for state in states:
                if state not in calc_states:
                    return "Invalid state provided: {}.".format(state)

        warnings_list = []

        now = timezone.now()

        if pks:
            q_object = Q(pk__in=pks)
        else:
            q_object = Q()

            if group is not None:
                g_pk = Group.get_from_string(group).pk
                q_object.add(Q(dbgroups__pk=g_pk), Q.AND)

            if group_pk is not None:
                q_object.add(Q(dbgroups__pk=group_pk), Q.AND)

            if not all_users:
                q_object.add(Q(user=get_automatic_user()), Q.AND)

            if states is not None:
                q_object.add(Q(dbattributes__key='state',
                               dbattributes__tval__in=states, ), Q.AND)
            if past_days is not None:
                now = timezone.now()
                n_days_ago = now - datetime.timedelta(days=past_days)
                q_object.add(Q(ctime__gte=n_days_ago), Q.AND)

        calc_list_pk = list(
            cls.query(q_object).distinct().values_list('pk', flat=True))

        calc_list = cls.query(pk__in=calc_list_pk).order_by('ctime')

        scheduler_states = dict(
            DbAttribute.objects.filter(dbnode__pk__in=calc_list_pk,
                                       key='scheduler_state').values_list(
                'dbnode__pk', 'tval'))

        # I do the query now, so that the list of pks gets cached
        calc_list_data = list(
            calc_list.filter(
                # dbcomputer__dbauthinfo__aiidauser=F('user')
            ).distinct().order_by('ctime').values(
                'pk', 'dbcomputer__name', 'ctime',
                'type', 'dbcomputer__enabled',
                'dbcomputer__pk',
                'user__pk'))
        list_comp_pk = [i['dbcomputer__pk'] for i in calc_list_data]
        list_aiduser_pk = [i['user__pk']
                           for i in calc_list_data]
        enabled_data = DbAuthInfo.objects.filter(
            dbcomputer__pk__in=list_comp_pk, aiidauser__pk__in=list_aiduser_pk
        ).values_list('dbcomputer__pk', 'aiidauser__pk', 'enabled')

        enabled_auth_dict = {(i[0], i[1]): i[2] for i in enabled_data}

        states = {c.pk: c._get_state_string() for c in calc_list}

        scheduler_lastcheck = dict(DbAttribute.objects.filter(
            dbnode__in=calc_list,
            key='scheduler_lastchecktime').values_list('dbnode__pk', 'dval'))

        ## Get the last daemon check
        try:
            last_daemon_check = get_last_daemon_timestamp('updater',
                                                          when='stop')
        except ValueError:
            last_check_string = ("# Last daemon state_updater check: "
                                 "(Error while retrieving the information)")
        else:
            if last_daemon_check is None:
                last_check_string = "# Last daemon state_updater check: (Never)"
            else:
                last_check_string = ("# Last daemon state_updater check: "
                                     "{} ({})".format(
                    str_timedelta(now - last_daemon_check,
                                  negative_to_zero=True),
                    timezone.localtime(last_daemon_check).strftime(
                        "at %H:%M:%S on %Y-%m-%d")))

        disabled_ignorant_states = [
            None, calc_states.FINISHED, calc_states.SUBMISSIONFAILED,
            calc_states.RETRIEVALFAILED, calc_states.PARSINGFAILED,
            calc_states.FAILED
        ]

        if not calc_list:
            return last_check_string
        else:
            # first save a matrix of results to be printed
            res_str_list = [last_check_string]
            str_matrix = []
            title = ['# Pk', 'State', 'Creation',
                     'Sched. state', 'Computer', 'Type']
            str_matrix.append(title)
            len_title = [len(i) for i in title]

            for calcdata in calc_list_data:
                remote_state = "None"

                calc_state = states[calcdata['pk']]
                remote_computer = calcdata['dbcomputer__name']
                try:
                    sched_state = scheduler_states.get(calcdata['pk'], None)
                    if sched_state is None:
                        remote_state = "(unknown)"
                    else:
                        remote_state = '{}'.format(sched_state)
                        if calc_state == calc_states.WITHSCHEDULER:
                            last_check = scheduler_lastcheck.get(calcdata['pk'],
                                                                 None)
                            if last_check is not None:
                                when_string = " {}".format(
                                    str_timedelta(now - last_check, short=True,
                                                  negative_to_zero=True))
                                verb_string = "was "
                            else:
                                when_string = ""
                                verb_string = ""
                            remote_state = "{}{}{}".format(verb_string,
                                                           sched_state,
                                                           when_string)
                except ValueError:
                    raise

                calc_module = \
                from_type_to_pluginclassname(calcdata['type']).rsplit(".", 1)[0]
                prefix = 'calculation.job.'
                prefix_len = len(prefix)
                if calc_module.startswith(prefix):
                    calc_module = calc_module[prefix_len:].strip()

                if relative_ctime:
                    calc_ctime = str_timedelta(now - calcdata['ctime'],
                                               negative_to_zero=True,
                                               max_num_fields=1)
                else:
                    calc_ctime = " ".join([timezone.localtime(
                        calcdata['ctime']).isoformat().split('T')[0],
                                           timezone.localtime(calcdata[
                                                                  'ctime']).isoformat().split(
                                               'T')[1].split('.')[
                                               0].rsplit(":", 1)[0]])

                the_state = states[calcdata['pk']]

                # decide if it is needed to print enabled/disabled information
                # By default, if the computer is not configured for the
                # given user, assume it is user_enabled
                user_enabled = enabled_auth_dict.get(
                    (calcdata['dbcomputer__pk'],
                     calcdata['user__pk']), True)
                global_enabled = calcdata["dbcomputer__enabled"]

                enabled = "" if (user_enabled and global_enabled or
                                 the_state in disabled_ignorant_states) else " [Disabled]"

                str_matrix.append([calcdata['pk'],
                                   the_state,
                                   calc_ctime,
                                   remote_state,
                                   remote_computer + "{}".format(enabled),
                                   calc_module
                                   ])

            # prepare a formatted text of minimal row length (to fit in terminals!)
            rows = []
            for j in range(len(str_matrix[0])):
                rows.append([len(str(i[j])) for i in str_matrix])
            line_lengths = [str(max(max(rows[i]), len_title[i])) for i in
                            range(len(rows))]
            fmt_string = "{:<" + "}|{:<".join(line_lengths) + "}"
            for row in str_matrix:
                res_str_list.append(fmt_string.format(*[str(i) for i in row]))

            res_str_list += ["# {}".format(_) for _ in warnings_list]
            return "\n".join(res_str_list)



# -*- coding: utf-8 -*-
"""
Testing the workfunction wrappers
"""

#from sqlalchemy.orm import sessionmaker, aliased
import time

from aiida.workflows2.specs.lib.wflib import *

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

@threadit
@workit_factory()
def top_func(self, **inp):
    self.test = True
    x = inp['x']
    print x.content
#    session4 = Session()
#    session4.add(self)
#    session4.add(x)
#    print "I am", self   # direct access to self will not work unless there is a session during execution
    print inp['x'].content
#    session4.close()
    er = IntData(5)
    print er.content['value']
    a = func10(self)
    b = func20(self, d=er)
    #c = a + b
    c = w4(a).update(w4(b))
    a1 = func10(self, e=er)
    a2 = func10(self, e=er)
    a3 = func10(self, e=er)
    a4 = func10(self)
    a5 = func10(self)

    #k = w4(a1)
    result = c
    print c
    return result

@threadit
@workit_factory()
def ex1(self):
    a = func10()
    b = func20()
    c = func10(w4(a))
    d = func10(w4(b))


@threadit
@workit_factory()
def func10(self, **inp):
    d = IntData(5)
    inp3 = {'a':IntData(3), 'b':d}
    result2 = w4(func20(self, **inp3))
    print inp3['a'].content['value']
    print result2
    time.sleep(1)
    return result2


@threadit
@workit_factory()
def func20(self, **inp):
    v= firstcomp(self, c=IntData(6))
    #result = StringData('result_2 ' + str(inp))
    result = {'f20res': StringData('result_3 ')}
    #print w4(v)
    #a = func10(self)
    #time.sleep(5)
    #print "---cdcd", inp['a'].content
    return result


@threadit
@workit_factory()
def firstcomp(self, c, d=IntData(8)):
    return {'first': IntData(6)}

#session5 = Session()
c = Calc(name='start')
x = IntData(88)
#session5.add_all([c, x])
#session5.commit()

top_func(c, x=x)
res = w4(top_func(c, x=x))
print "result {0}".format(res)
#for item in res:
#    item.roots.append(x)
#session5.commit()

#*******************************

@threadit
@workit_factory()
def func0(self, **inp):
    N = inp['x']
    M = inp['y']
    K = IntData(37)
    output = {'N':N, 'M':M, 'K':K}
    time.sleep(5)
    return output


@threadit
@workit_factory()
def func1(self, **inp):
    n = inp['number']
    double = n * 2
    triple = n * 3
    output = {'outA':double, 'outB':triple}
    time.sleep(n)
    return output


@threadit
@workit_factory()
def func2(self, **inp):
    n = inp['number']
    square = n * n
    cube = n * n * n
    output = {'outC':square, 'outD':cube}
    time.sleep(n)
    return output


@threadit
@workit_factory()
def func3(self, inE, inF):
    #e = inp['inE']
    #f = inp['inF']
    e = inE
    f = inF
    output= {'combined':e + f}
    return output


@threadit
@workit_factory()
def Workflow3(self, **inp):
    step0 = func0(self, x=inp['first'], y=inp['second'])
    result0 = w4(step0)

    step1 = func1(self, number=result0['N'])  #these two are started in parallel
    step2 = func2(self, number=result0['M'])

    result1 = w4(step1)
    result2 = w4(step2)   # block and wait for both results before proceeding

    step3 = func3(self, inE=result1['outB'], inF=result2['outC'])
    result3 = w4(step3)
    output = {'subsum':result3['combined']}
    return output

@threadit
@workit_factory()
def collect4(self, arr):
    output = {'sum': sum(arr)}
    return output

@threadit
@workit_factory()
def Workflow4(self, **inp):
    step = []   #initiate lists
    result = []
    N = inp['Nmax']

    # start multiple parallel runs
    for n in xrange(N):
        step.append(Workflow3(self, first=n, second=n))

    # collect the results once they are all finished
    for n in xrange(N):
        x = w4(step[n])
        result.append(x['subsum'])

    print result
    output = collect4(self, arr = tuple(result))   #tuple to make hashable value
    return w4(output)


#b = Workflow4(c, Nmax=IntData(5))
#print w4(b)

#b = Workflow4(Wstart, Nmax=4)
#print w4(b)



# -*- coding: utf-8 -*-
import os
import sys
import traceback
import inspect
import logging

from abc import abstractmethod, abstractproperty

from aiida.common.exceptions import (InternalError, AiidaException,
                                     InvalidOperation, RemoteOperationError)
from aiida.common.datastructures import (wf_states, wf_exit_call,
                                         wf_default_call, calc_states)
from aiida.common.utils import str_timedelta
from aiida.common import aiidalogger
from aiida.orm.implementation.calculation import JobCalculation



from aiida.backends.utils import get_automatic_user

from aiida.utils import timezone
from aiida.utils.logger import get_dblogger_extra

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

logger = aiidalogger.getChild('Workflow')


class WorkflowKillError(AiidaException):
    """
    An exception raised when a workflow failed to be killed.
    The error_message_list attribute contains the error messages from
    all the subworkflows.
    """

    def __init__(self, *args, **kwargs):
        # Call the base class constructor with the parameters it needs
        super(WorkflowKillError, self).__init__(*args)

        self.error_message_list = kwargs.pop('error_message_list', [])
        if kwargs:
            raise ValueError("Unknown parameters passed to WorkflowKillError")


class WorkflowUnkillable(AiidaException):
    """
    Raised when a workflow cannot be killed because it is in the FINISHED or
    ERROR state.
    """
    pass


class AbstractWorkflow(object):
    """
    Base class to represent a workflow. This is the superclass of any workflow implementations,
    and provides all the methods necessary to interact with the database.

    The typical use case are workflow stored in the aiida.workflow packages, that are initiated
    either by the user in the shell or by some scripts, and that are monitored by the aiida daemon.

    Workflow can have steps, and each step must contain some calculations to be executed. At the
    end of the step's calculations the workflow is reloaded in memory and the next methods is called.

    .. todo: verify if there are other places (beside label and description) where
      the _increment_version_number_db routine needs to be called to increase
      the nodeversion after storing
    """
    # Name to be used for the repository section
    _section_name = 'workflow'

    # The name of the subfolder in which to put the files/directories added with add_path
    _path_subfolder_name = 'path'

    @abstractmethod
    def __init__(self, **kwargs):
        """
        Initializes the Workflow super class, store the instance in the DB and in case
        stores the starting parameters.

        If initialized with an uuid the Workflow is loaded from the DB, if not a new
        workflow is generated and added to the DB following the stack frameworks. This
        means that only modules inside aiida.workflows are allowed to implements
        the workflow super calls and be stored. The caller names, modules and files are
        retrieved from the stack.

        :param uuid: a string with the uuid of the object to be loaded.
        :param params: a dictionary of storable objects to initialize the specific workflow
        :raise: NotExistent: if there is no entry of the desired workflow kind with
                             the given uuid.
        """
        pass

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, str(self))

    def __str__(self):
        if self._to_be_stored:
            return "uuid: {} (unstored)".format(self.uuid)
        else:
            return "uuid: {} (pk: {})".format(self.uuid, self.pk)


    @abstractproperty
    def dbworkflowinstance(self):
        """
        Get the DbWorkflow object stored in the super class.

        :return: DbWorkflow object from the database
        """
        pass

    @abstractmethod
    def _get_dbworkflowinstance(self):
        pass

    @abstractproperty
    def label(self):
        """
        Get the label of the workflow.

        :return: a string
        """
        pass

    @label.setter
    @abstractmethod
    def label(self, label):
        """
        Set the label of the workflow.

        :param label: a string
        """
        pass

    @abstractmethod
    def _update_db_label_field(self, field_value):
        """
        Safety method to store the label of the workflow

        :return: a string
        """
        pass

    @abstractproperty
    def description(self):
        """
        Get the description of the workflow.

        :return: a string
        """
        pass

    @description.setter
    @abstractmethod
    def description(self, desc):
        """
        Set the description of the workflow

        :param desc: a string
        """
        pass

    @abstractmethod
    def _update_db_description_field(self, field_value):
        """
        Safety method to store the description of the workflow

        :return: a string
        """
        pass


    @abstractmethod
    def _increment_version_number_db(self):
        """
        This function increments the version number in the DB.
        This should be called every time you need to increment the version (e.g. on adding a
        extra or attribute).
        """
        pass

    @property
    def repo_folder(self):
        """
        Get the permanent repository folder.
        Use preferentially the current_folder method.

        :return: the permanent RepositoryFolder object
        """
        return self._repo_folder

    @property
    def current_folder(self):
        """
        Get the current repository folder,
        whether the temporary or the permanent.

        :return: the RepositoryFolder object.
        """
        if self._to_be_stored:
            return self.get_temp_folder()
        else:
            return self.repo_folder

    @property
    def _get_folder_pathsubfolder(self):
        """
        Get the subfolder in the repository.

        :return: a Folder object.
        """
        return self.current_folder.get_subfolder(
            self._path_subfolder_name, reset_limit=True)

    def get_folder_list(self, subfolder='.'):
        """
        Get the the list of files/directory in the repository of the object.

        :param str,optional subfolder: get the list of a subfolder
        :return: a list of strings.
        """
        return self._get_folder_pathsubfolder.get_subfolder(subfolder).get_content_list()

    def get_temp_folder(self):
        """
        Get the folder of the Node in the temporary repository.

        :return: a SandboxFolder object mapping the node in the repository.
        """
        if self._temp_folder is None:
            raise InternalError("The temp_folder was asked for node {}, but it is "
                                "not set!".format(self.uuid))
        return self._temp_folder

    def remove_path(self, path):
        """
        Remove a file or directory from the repository directory.

        Can be called only before storing.
        """
        if not self._to_be_stored:
            raise ValueError("Cannot delete a path after storing the node")

        if os.path.isabs(path):
            raise ValueError("The destination path in remove_path must be a relative path")
        self._get_folder_pathsubfolder.remove_path(path)

    def add_path(self, src_abs, dst_path):
        """
        Copy a file or folder from a local file inside the repository directory.
        If there is a subpath, folders will be created.

        Copy to a cache directory if the entry has not been saved yet.
        src_abs: the absolute path of the file to copy.
        dst_filename: the (relative) path on which to copy.
        """
        if not self._to_be_stored:
            raise ValueError("Cannot insert a path after storing the node")

        if not os.path.isabs(src_abs):
            raise ValueError("The source path in add_path must be absolute")
        if os.path.isabs(dst_path):
            raise ValueError("The destination path in add_path must be a filename without any subfolder")
        self._get_folder_pathsubfolder.insert_path(src_abs, dst_path)

    def get_abs_path(self, path, section=None):
        """
        TODO: For the moment works only for one kind of files, 'path' (internal files)
        """
        if section is None:
            section = self._path_subfolder_name

        if os.path.isabs(path):
            raise ValueError("The path in get_abs_path must be relative")
        return self.current_folder.get_subfolder(section, reset_limit=True).get_abs_path(path, check_existence=True)

    @classmethod
    @abstractmethod
    def query(cls, *args, **kwargs):
        """
        Map to the aiidaobjects manager of the DbWorkflow, that returns
        Workflow objects instead of DbWorkflow entities.

        """
        pass

    @property
    def logger(self):
        """
        Get the logger of the Workflow object, so that it also logs to the
        DB.

        :return: LoggerAdapter object, that works like a logger, but also has
          the 'extra' embedded
        """
        return logging.LoggerAdapter(logger=self._logger,
                                     extra=get_dblogger_extra(self))

    @abstractmethod
    def store(self):
        """
        Stores the DbWorkflow object data in the database
        """
        pass

    @abstractproperty
    def uuid(self):
        """
        Returns the DbWorkflow uuid
        """
        pass

    @abstractproperty
    def pk(self):
        """
        Returns the DbWorkflow pk
        """
        pass

    @abstractmethod
    def info(self):
        """
        Returns an array with all the informations about the modules, file, class to locate
        the workflow source code
        """
        pass

    @abstractmethod
    def set_params(self, params, force=False):
        """
        Adds parameters to the Workflow that are both stored and used every time
        the workflow engine re-initialize the specific workflow to launch the new methods.
        """
        pass

    @abstractmethod
    def get_parameters(self):
        """
        Get the Workflow paramenters
        :return: a dictionary of storable objects
        """
        pass

    @abstractmethod
    def get_parameter(self, _name):
        """
        Get one Workflow paramenter
        :param name: a string with the parameters name to retrieve
        :return: a dictionary of storable objects
        """
        pass

    @abstractmethod
    def get_attributes(self):
        """
        Get the Workflow attributes
        :return: a dictionary of storable objects
        """
        pass

    @abstractmethod
    def get_attribute(self, _name):
        """
        Get one Workflow attribute
        :param name: a string with the attribute name to retrieve
        :return: a dictionary of storable objects
        """
        pass

    @abstractmethod
    def add_attributes(self, _params):
        """
        Add a set of attributes to the Workflow. If another attribute is present with the same name it will
        be overwritten.
        :param name: a string with the attribute name to store
        :param value: a storable object to store
        """
        pass

    @abstractmethod
    def add_attribute(self, _name, _value):
        """
        Add one attributes to the Workflow. If another attribute is present with the same name it will
        be overwritten.
        :param name: a string with the attribute name to store
        :param value: a storable object to store
        """
        pass

    @abstractmethod
    def get_results(self):
        """
        Get the Workflow results
        :return: a dictionary of storable objects
        """
        pass

    @abstractmethod
    def get_result(self, _name):
        """
        Get one Workflow result
        :param name: a string with the result name to retrieve
        :return: a dictionary of storable objects
        """
        pass

    @abstractmethod
    def add_results(self, _params):
        """
        Add a set of results to the Workflow. If another result is present with the same name it will
        be overwritten.
        :param name: a string with the result name to store
        :param value: a storable object to store
        """
        pass

    @abstractmethod
    def add_result(self, _name, _value):
        """
        Add one result to the Workflow. If another result is present with the same name it will
        be overwritten.
        :param name: a string with the result name to store
        :param value: a storable object to store
        """
        pass

    @abstractmethod
    def get_state(self):
        """
        Get the Workflow's state
        :return: a state from wf_states in aiida.common.datastructures
        """
        pass

    @abstractmethod
    def set_state(self, state):
        """
        Set the Workflow's state
        :param name: a state from wf_states in aiida.common.datastructures
        """
        pass

    @abstractmethod
    def is_new(self):
        """
        Returns True is the Workflow's state is CREATED
        """
        pass

    @abstractmethod
    def is_running(self):
        """
        Returns True is the Workflow's state is RUNNING
        """
        pass

    @abstractmethod
    def has_finished_ok(self):
        """
        Returns True is the Workflow's state is FINISHED
        """
        pass

    @abstractmethod
    def has_failed(self):
        """
        Returns True is the Workflow's state is ERROR
        """
        pass

    @abstractmethod
    def is_subworkflow(self):
        """
        Return True is this is a subworkflow (i.e., if it has a parent),
        False otherwise.
        """
        pass

    @abstractmethod
    def get_step(self, step_method):

        """
        Retrieves by name a step from the Workflow.
        :param step_method: a string with the name of the step to retrieve or a method
        :raise: ObjectDoesNotExist: if there is no step with the specific name.
        :return: a DbWorkflowStep object.
        """
        pass

    @abstractmethod
    def get_steps(self, state=None):
        """
        Retrieves all the steps from a specific workflow Workflow with the possibility to limit the list
        to a specific step's state.
        :param state: a state from wf_states in aiida.common.datastructures
        :return: a list of DbWorkflowStep objects.
        """
        pass

    def has_step(self, step_method):
        """
        Return if the Workflow has a step with a specific name.
        :param step_method: a string with the name of the step to retrieve or a method
        """
        return not self.get_step(step_method) == None

    @classmethod
    def step(cls, fun):
        """
        This method is used as a decorator for workflow steps, and handles the method's execution,
        the state updates and the eventual errors.

        The decorator generates a wrapper around the input function to execute, adding with the correct
        step name and a utility variable to make it distinguishable from non-step methods.

        When a step is launched, the wrapper tries to run the function in case of error the state of
        the workflow is moved to ERROR and the traceback is stored in the report. In general the input
        method is a step obtained from the Workflow object, and the decorator simply handles a controlled
        execution of the step allowing the code not to break in case of error in the step's source code.

        The wrapper also tests not to run two times the same step, unless a Workflow is in ERROR state, in this
        case all the calculations and subworkflows of the step are killed and a new execution is allowed.

        :param fun: a methods to wrap, making it a Workflow step
        :raise: AiidaException: in case the workflow state doesn't allow the execution
        :return: the wrapped methods,
        """
        wrapped_method = fun.__name__

        # This function gets called only if the method is launched with the execution brackets ()
        # Otherwise, when the method is addressed in a next() call this never gets called and only the
        # attributes are added
        def wrapper(cls, *args, **kwargs):
            # Store the workflow at the first step executed
            if cls._to_be_stored:
                cls.store()

            if len(args) > 0:
                raise AiidaException("A step method cannot have any argument, use add_attribute to the workflow")

            # If a method is launched and the step is RUNNING or INITIALIZED we should stop
            if cls.has_step(wrapped_method) and \
                    not (cls.get_step(wrapped_method).state == wf_states.ERROR or \
                                     cls.get_step(wrapped_method).state == wf_states.SLEEP or \
                                     cls.get_step(wrapped_method).nextcall == wf_default_call or \
                                     cls.get_step(wrapped_method).nextcall == wrapped_method \
                         #cls.has_step(wrapped_method) \
                    ):
                raise AiidaException(
                    "The step {0} has already been initialized, cannot change this outside the parent workflow !".format(
                        wrapped_method))

            # If a method is launched and the step is halted for ERROR, then clean the step and re-launch
            if cls.has_step(wrapped_method) and \
                    ( cls.get_step(wrapped_method).state == wf_states.ERROR or \
                                  cls.get_step(wrapped_method).state == wf_states.SLEEP ):

                for w in cls.get_step(wrapped_method).get_sub_workflows(): w.kill()
                cls.get_step(wrapped_method).remove_sub_workflows()

                for c in cls.get_step(wrapped_method).get_calculations(): c.kill()
                cls.get_step(wrapped_method).remove_calculations()

                #self.get_steps(wrapped_method).set_nextcall(wf_exit_call)

            method_step, created = cls.dbworkflowinstance.steps.get_or_create(name=wrapped_method,
                                                                              user=get_automatic_user())

            try:
                fun(cls)
            except:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                cls.append_to_report(
                    "ERROR ! This workflow got an error in the {0} method, we report down the stack trace".format(
                        wrapped_method))
                cls.append_to_report("full traceback: {0}".format(traceback.format_exc()))
                method_step.set_state(wf_states.ERROR)
            return None

        out = wrapper
        wrapper.is_wf_step = True
        wrapper.wf_step_name = fun.__name__

        return wrapper

    def next(self, next_method):
        """
        Adds the a new step to be called after the completion of the caller method's calculations and subworkflows.

        This method must be called inside a Workflow step, otherwise an error is thrown. The
        code finds the caller method and stores in the database the input next_method as the next
        method to be called. At this point no execution in made, only configuration updates in the database.

        If during the execution of the caller method the user launched calculations or subworkflows, this
        method will add them to the database, making them available to the workflow manager to be launched.
        In fact all the calculation and subworkflow submissions are lazy method, really executed by this call.

        :param next_method: a Workflow step method to execute after the caller method
        :raise: AiidaException: in case the caller method cannot be found or validated
        :return: the wrapped methods, decorated with the correct step name
        """

        md5 = self.dbworkflowinstance.script_md5
        script_path = self.dbworkflowinstance.script_path

        # TODO: in principles, the file containing the workflow description
        # should be copied in a repository, and, on that file, the workflow
        # should check to be sure of loading the same description of the
        # workflow. At the moment, this is not done and is checking the source
        # in aiida/workflows/... resulting essentially in the impossibility of
        # developing a workflow without rendering most of the trial run
        # unaccessible. I comment these lines for this moment.

        #if md5 != md5_file(script_path):
        #    raise ValidationError("Unable to load the original workflow module from {}, MD5 has changed".format(script_path))

        # ATTENTION: Do not move this code outside or encapsulate it in a function
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        caller_method = calframe[1][3]

        if next_method is None:
            raise AiidaException("The next method is None, probably you passed a method with parenthesis ??")

        if not self.has_step(caller_method):
            raise AiidaException("The caller method is either not a step or has not been registered as one")

        if not next_method.__name__ == wf_exit_call:
            try:
                is_wf_step = getattr(next_method, "is_wf_step", None)
            except AttributeError:
                raise AiidaException("Cannot add as next call a method not decorated as Workflow method")

        # TODO SP: abstract this, this depends on the DB. The better would be
        # to add a method to the DbWorkflow from SQLA and Django to get steps
        # with particular filters, in order to avoid repetition of all the code
        # arround

        # Retrieve the caller method
        method_step = self.dbworkflowinstance.steps.get(name=caller_method, user=get_automatic_user())

        # Attach calculations
        if caller_method in self.attach_calc_lazy_storage:
            for c in self.attach_calc_lazy_storage[caller_method]:
                method_step.add_calculation(c)

        # Attach sub-workflows
        if caller_method in self.attach_subwf_lazy_storage:
            for w in self.attach_subwf_lazy_storage[caller_method]:
                method_step.add_sub_workflow(w)

        # Set the next method
        if not next_method.__name__ == wf_exit_call:
            next_method_name = next_method.wf_step_name
        else:
            next_method_name = wf_exit_call

        #logger.info("Adding step {0} after {1} in {2}".format(next_method_name, caller_method, self.uuid))
        method_step.set_nextcall(next_method_name)
        #
        self.dbworkflowinstance.set_state(wf_states.RUNNING)
        method_step.set_state(wf_states.RUNNING)

    def attach_calculation(self, calc):
        """
        Adds a calculation to the caller step in the database. This is a lazy call, no
        calculations will be launched until the ``next`` method gets called. For a step to be
        completed all the calculations linked have to be in RETRIEVED state, after which the next
        method gets called from the workflow manager.
        :param calc: a JobCalculation object
        :raise: AiidaException: in case the input is not of JobCalculation type
        """

        if (not issubclass(calc.__class__, JobCalculation) and not isinstance(calc, JobCalculation)):
            raise AiidaException("Cannot add a calculation not of type JobCalculation")

        for node in calc.get_inputs():
            if node.pk is None:
                raise AiidaException("Cannot add a calculation with "
                                     "unstored inputs")

            #        if calc.pk is None:
            #            raise AiiDAException("Cannot add an unstored calculation")

        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        caller_funct = calframe[1][3]

        if not caller_funct in self.attach_calc_lazy_storage:
            self.attach_calc_lazy_storage[caller_funct] = []
        self.attach_calc_lazy_storage[caller_funct].append(calc)

    def attach_workflow(self, sub_wf):
        """
        Adds a workflow to the caller step in the database. This is a lazy call, no
        workflow will be started until the ``next`` method gets called. For a step to be
        completed all the workflows linked have to be in FINISHED state, after which the next
        method gets called from the workflow manager.
        :param next_method: a Workflow object
        """
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        caller_funct = calframe[1][3]

        if not caller_funct in self.attach_subwf_lazy_storage:
            self.attach_subwf_lazy_storage[caller_funct] = []
        self.attach_subwf_lazy_storage[caller_funct].append(sub_wf)

    def get_step_calculations(self, step_method, calc_state=None):
        """
        Retrieves all the calculations connected to a specific step in the database. If the step
        is not existent it returns None, useful for simpler grammatic in the workflow definition.
        :param next_method: a Workflow step (decorated) method
        :param calc_state: a specific state to filter the calculations to retrieve
        :return: a list of JobCalculations objects
        """
        if not getattr(step_method, "is_wf_step"):
            raise AiidaException("Cannot get step calculations from a method not decorated as Workflow method")

        step_method_name = step_method.wf_step_name

        try:
            stp = self.get_step(step_method_name)
            return stp.get_calculations(state=calc_state)
        except:
            raise AiidaException("Cannot retrieve step's calculations")

    def get_step_workflows(self, step_method):
        """
        Retrieves all the workflows connected to a specific step in the database. If the step
        is not existent it returns None, useful for simpler grammatic in the workflow definition.
        :param next_method: a Workflow step (decorated) method
        """
        if not getattr(step_method, "is_wf_step"):
            raise AiidaException("Cannot get step calculations from a method not decorated as Workflow method")

        step_method_name = step_method.wf_step_name

        stp = self.get_step(step_method_name)
        return stp.get_sub_workflows()

    def kill_step_calculations(self, step):
        """
        Calls the ``kill`` method for each Calculation linked to the step method passed as argument.
        :param step: a Workflow step (decorated) method
        """
        counter = 0
        for c in step.get_calculations():
            if c._is_new() or c._is_running():
                try:
                    c.kill()
                except (InvalidOperation, RemoteOperationError) as e:
                    counter += 1
                    self.logger.error(e.message)

        if counter:
            raise InvalidOperation("{} step calculation{} could not be killed"
                                   "".format(counter, "" if counter == 1 else "s"))

    def kill(self, verbose=False):
        """
        Stop the Workflow execution and change its state to FINISHED.

        This method calls the ``kill`` method for each Calculation and each
        subworkflow linked to each RUNNING step.

        :param verbose: True to print the pk of each subworkflow killed
        :raise InvalidOperation: if some calculations cannot be killed (the
                                 workflow will be also put to SLEEP so that it
                                 can be killed later on)
        """
        if self.get_state() not in [wf_states.FINISHED, wf_states.ERROR]:

            # put in SLEEP state first
            self.dbworkflowinstance.set_state(wf_states.SLEEP)

            error_messages = []
            for s in self.get_steps(state=wf_states.RUNNING):

                try:
                    self.kill_step_calculations(s)
                except InvalidOperation as e:
                    error_message = ("'{}' for workflow with pk= {}"
                                     "".format(e.message, self.pk))
                    error_messages.append(error_message)

                for w in s.get_sub_workflows():
                    if verbose:
                        print "Killing workflow with pk: {}".format(w.pk)

                    try:
                        w.kill(verbose=verbose)
                    except WorkflowKillError as e:
                        #self.logger.error(e.message)
                        error_messages.extend(e.error_message_list)
                    except WorkflowUnkillable:
                        # A subwf cannot be killed, skip
                        pass

            if self.get_steps(state=wf_states.CREATED):
                error_messages.append("Workflow with pk= {} cannot be killed "
                                      "because some steps are in {} state"
                                      "".format(self.pk, wf_states.CREATED))

            if error_messages:
                raise WorkflowKillError("Workflow with pk= {} cannot be "
                                        "killed and was put to {} state instead; "
                                        "try again later".format(self.pk, wf_states.SLEEP),
                                        error_message_list=error_messages)
            else:
                self.dbworkflowinstance.set_state(wf_states.FINISHED)
        else:
            raise WorkflowUnkillable("Cannot kill a workflow in {} or {} state"
                                     "".format(wf_states.FINISHED, wf_states.ERROR))

    def get_all_calcs(self, calc_class=JobCalculation,calc_state=None,depth=15):
        """
        Get all calculations connected with this workflow and all its subworflows up to a given depth.
        The list of calculations can be restricted to a given calculation type and state
        :param calc_class: the calculation class to which the calculations should belong (default: JobCalculation)

        :param calc_state: a specific state to filter the calculations to retrieve

        :param depth: the maximum depth level the recursion on sub-workflows will
          try to reach (0 means we stay at the step level and don't go
          into sub-workflows, 1 means we go down to one step level of
          the sub-workflows, etc.)

        :return: a list of JobCalculation objects
        """

        all_calcs = []
        for st in self.get_steps():
            all_calcs += [ c for c in st.get_calculations(state=calc_state) if isinstance(c,calc_class)]
            if depth>0:
                for subw in st.get_sub_workflows():
                    all_calcs += subw.get_all_calcs(calc_state=calc_state,calc_class=calc_class,depth=depth-1)
        return all_calcs

    def sleep(self):
        """
        Changes the workflow state to SLEEP, only possible to call from a Workflow step decorated method.
        """
        # ATTENTION: Do not move this code outside or encapsulate it in a function
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        caller_method = calframe[1][3]

        if not self.has_step(caller_method):
            raise AiidaException("The caller method is either not a step or has not been registered as one")

        self.get_step(caller_method).set_state(wf_states.SLEEP)

    @abstractmethod
    def get_report(self):
        """
        Return the Workflow report.

        :note: once, in case the workflow is a subworkflow of any other Workflow this method
          calls the parent ``get_report`` method.
          This is not the case anymore.
        :return: a list of strings
        """
        pass

    @abstractmethod
    def clear_report(self):
        """
        Wipe the Workflow report. In case the workflow is a subworflow of any other Workflow this method
        calls the parent ``clear_report`` method.
        """
        pass

    @abstractmethod
    def append_to_report(self, text):
        """
        Adds text to the Workflow report.

        :note: Once, in case the workflow is a subworkflow of any other Workflow this method
         calls the parent ``append_to_report`` method; now instead this is not the
         case anymore
        """
        pass

    @classmethod
    @abstractmethod
    def get_subclass_from_dbnode(cls, wf_db):
        """
        Loads the workflow object and reaoads the python script in memory with the importlib library, the
        main class is searched and then loaded.
        :param wf_db: a specific DbWorkflowNode object representing the Workflow
        :return: a Workflow subclass from the specific source code
        """
        pass

    @classmethod
    @abstractmethod
    def get_subclass_from_pk(cls, pk):
        """
        Calls the ``get_subclass_from_dbnode`` selecting the DbWorkflowNode from
        the input pk.
        :param pk: a primary key index for the DbWorkflowNode
        :return: a Workflow subclass from the specific source code
        """
        pass

    @classmethod
    @abstractmethod
    def get_subclass_from_uuid(cls, uuid):
        """
        Calls the ``get_subclass_from_dbnode`` selecting the DbWorkflowNode from
        the input uuid.
        :param uuid: a uuid for the DbWorkflowNode
        :return: a Workflow subclass from the specific source code
        """
        pass

    def exit(self):
        """
        This is the method to call in  ``next`` to finish the Workflow. When exit is the next method,
        and no errors are found, the Workflow is set to FINISHED and removed from the execution manager
        duties.
        """
        pass


# def revive(self):
#
#         from aiida.common.utils import md5_file
#         md5 = self.dbworkflowinstance.script_md5
#         script_path = self.dbworkflowinstance.script_path
#
#         md5_check = md5_file(script_path)
#
#         # MD5 Check before revive
#         if md5 != md5_check:
#             logger.info("The script has changed, MD5 is now updated")
#             self.dbworkflowinstance.set_script_md5(md5_check)
#
#         # Clear all the erroneous steps
#         err_steps    = self.get_steps(state=wf_states.ERROR)
#         for s in err_steps:
#
#             for w in s.get_sub_workflows(): w.kill()
#             s.remove_sub_workflows()
#
#             for c in s.get_calculations(): c.kill()
#             s.remove_calculations()
#
#             s.set_state(wf_states.INITIALIZED)
#
#         self.set_state(wf_states.RUNNING)



def kill_all():
    """
    Kills all the workflows not in FINISHED state running the ``kill_from_uuid``
    method in a loop.

    :param uuid: the UUID of the workflow to kill
    """

    raise NotImplementedError

def kill_from_pk():
    """
    Kills a workflow from its pk.

    :param pk: the Pkof the workflow to kill
    """

    raise NotImplementedError


def get_workflow_info(w, tab_size=2, short=False, pre_string="",
                      depth=16):
    """
    Return a string with all the information regarding the given workflow and
    all its calculations and subworkflows.
    This is a recursive function (to print all subworkflows info as well).

    :param w: a DbWorkflow instance
    :param tab_size: number of spaces to use for the indentation
    :param short: if True, provide a shorter output (only total number of
        calculations, rather than the state of each calculation)
    :param pre_string: string appended at the beginning of each line
    :param depth: the maximum depth level the recursion on sub-workflows will
                  try to reach (0 means we stay at the step level and don't go
                  into sub-workflows, 1 means we go down to one step level of
                  the sub-workflows, etc.)

    :return lines: list of lines to be outputed
    """
    # Note: pre_string becomes larger at each call of get_workflow_info on the
    #       subworkflows: pre_string -> pre_string + "|" + " "*(tab_size-1))
    # TODO SP: abstract the dependence on DbWorkflow

    from aiida.backends.djsite.db.models import DbWorkflow

    if tab_size < 2:
        raise ValueError("tab_size must be > 2")

    now = timezone.now()

    lines = []

    if w.label:
        wf_labelstring = "'{}', ".format(w.label)
    else:
        wf_labelstring = ""

    lines.append(pre_string)  # put an empty line before any workflow
    lines.append(pre_string + "+ Workflow {} ({}pk: {}) is {} [{}]".format(
        w.module_class, wf_labelstring, w.pk, w.state, str_timedelta(
            now - w.ctime, negative_to_zero=True)))

    # print information on the steps only if depth is higher than 0
    if depth > 0:

        # order all steps by time and  get all the needed values
        steps_and_subwf_pks = w.steps.all().order_by('time', 'sub_workflows__ctime',
                                                     'calculations__ctime').values_list('pk',
                                                                                        'sub_workflows__pk',
                                                                                        'calculations', 'name',
                                                                                        'nextcall', 'state')
        # get the list of step pks (distinct), preserving the order
        steps_pk = []
        for item in steps_and_subwf_pks:
            if item[0] not in steps_pk:
                steps_pk.append(item[0])

        # build a dictionary with all the infos for each step pk
        subwfs_of_steps = {}
        for step_pk, subwf_pk, calc_pk, name, nextcall, state in steps_and_subwf_pks:
            if step_pk not in subwfs_of_steps.keys():
                subwfs_of_steps[step_pk] = {'name': name,
                                            'nextcall': nextcall,
                                            'state': state,
                                            'subwf_pks': [],
                                            'calc_pks': [],
                }
            if subwf_pk:
                subwfs_of_steps[step_pk]['subwf_pks'].append(subwf_pk)
            if calc_pk:
                subwfs_of_steps[step_pk]['calc_pks'].append(calc_pk)


        # TODO SP: abstract this
        # get all subworkflows for all steps
        wflows = DbWorkflow.objects.filter(parent_workflow_step__in=steps_pk)  #.order_by('ctime')
        # dictionary mapping pks into workflows
        workflow_mapping = {_.pk: _ for _ in wflows}

        # get all calculations for all steps
        calcs = JobCalculation.query(workflow_step__in=steps_pk)  #.order_by('ctime')
        # dictionary mapping pks into calculations
        calc_mapping = {_.pk: _ for _ in calcs}

        for step_pk in steps_pk:
            lines.append(pre_string + "|" + '-' * (tab_size - 1) +
                         "* Step: {0} [->{1}] is {2}".format(
                             subwfs_of_steps[step_pk]['name'],
                             subwfs_of_steps[step_pk]['nextcall'],
                             subwfs_of_steps[step_pk]['state']))

            calc_pks = subwfs_of_steps[step_pk]['calc_pks']

            # print calculations only if it is not short
            if short:
                lines.append(pre_string + "|" + " " * (tab_size - 1) +
                             "| [{0} calculations]".format(len(calc_pks)))
            else:
                for calc_pk in calc_pks:
                    c = calc_mapping[calc_pk]
                    calc_state = c.get_state()
                    if c.label:
                        labelstring = "'{}', ".format(c.label)
                    else:
                        labelstring = ""

                    if calc_state == calc_states.WITHSCHEDULER:
                        sched_state = c.get_scheduler_state()
                        if sched_state is None:
                            remote_state = "(remote state still unknown)"
                        else:
                            last_check = c._get_scheduler_lastchecktime()
                            if last_check is not None:
                                when_string = " {}".format(
                                    str_timedelta(now - last_check, short=True,
                                                  negative_to_zero=True))
                                verb_string = "was "
                            else:
                                when_string = ""
                                verb_string = ""
                            remote_state = " ({}{}{})".format(verb_string,
                                                              sched_state, when_string)
                    else:
                        remote_state = ""
                    lines.append(pre_string + "|" + " " * (tab_size - 1) +
                                 "| Calculation ({}pk: {}) is {}{}".format(
                                     labelstring, calc_pk, calc_state, remote_state))

            ## SubWorkflows
            for subwf_pk in subwfs_of_steps[step_pk]['subwf_pks']:
                subwf = workflow_mapping[subwf_pk]
                lines.extend(get_workflow_info(subwf,
                                               short=short, tab_size=tab_size,
                                               pre_string=pre_string + "|" + " " * (tab_size - 1),
                                               depth=depth - 1))

            lines.append(pre_string + "|")

    return lines




# -*- coding: utf-8 -*-
from __future__ import absolute_import

from sqlalchemy import orm
from sqlalchemy.orm.exc import UnmappedClassError
from sqlalchemy.ext.declarative import declarative_base

from aiida.backends import sqlalchemy as sa
from aiida.common.exceptions import InvalidOperation

# Taken from
# https://github.com/mitsuhiko/flask-sqlalchemy/blob/master/flask_sqlalchemy/__init__.py#L491


__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

class _QueryProperty(object):

    def __init__(self, query_class=orm.Query):
        self.query_class = query_class

    def __get__(self, obj, _type):
        try:
            mapper = orm.class_mapper(_type)
            if mapper:
                return self.query_class(mapper, session=sa.session)
        except UnmappedClassError:
            return None


class _SessionProperty(object):
    def __get__(self, obj, _type):
        if not sa.session:
            raise InvalidOperation("You need to call load_dbenv before "
                                   "accessing the session of SQLALchemy.")
        return sa.session

class _AiidaQuery(orm.Query):

    def __init__(self, *args, **kwargs):
            super(_AiidaQuery, self).__init__(*args, **kwargs)

    def __iter__(self):
        iterator = super(_AiidaQuery, self).__iter__()
        for r in iterator:
            # Allow the use of with_entities
            if issubclass(type(r), Model):
                yield r.get_aiida_class()
            else:
                yield r


class Model(object):

    query = _QueryProperty()

    session = _SessionProperty()

    def save(self, commit=True):
        self.session.add(self)
        if commit:
            self.session.commit()
        return self

    def delete(self, commit=True):
        self.session.delete(self)
        if commit:
            self.session.commit()

Base = declarative_base(cls=Model, name='Model')



# -*- coding: utf-8 -*-
import aiida.common
from aiida.common.exceptions import InternalError
from aiida.common.extendeddicts import FixedFieldsAttributeDict

import os, re, fnmatch, sys  # for glob commands

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. and 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014 Python Software Foundation. All rights reserved."
__license__ = "MIT license, and Python license, see LICENSE.txt file"
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

magic_check = re.compile('[*?[]')


def TransportFactory(module):
    """
    Used to return a suitable Transport subclass.

    :param str module: name of the module containing the Transport subclass
    :return: the transport subclass located in module 'module'
    """
    from aiida.common.pluginloader import BaseFactory

    return BaseFactory(module, Transport, "aiida.transport.plugins")


class FileAttribute(FixedFieldsAttributeDict):
    """
    A class, resembling a dictionary, to describe the attributes of a file,
    that is returned by get_attribute().
    Possible keys: st_size, st_uid, st_gid, st_mode, st_atime, st_mtime
    """
    _valid_fields = (
        'st_size',
        'st_uid',
        'st_gid',
        'st_mode',
        'st_atime',
        'st_mtime',
    )


class TransportInternalError(InternalError):
    """
    Raised if there is a transport error that is raised to an internal error (e.g.
    a transport method called without opening the channel first).
    """
    pass


def copy_from_remote_to_remote(transportsource,transportdestination,
                                  remotesource,remotedestination,**kwargs):
    """
    Copy files or folders from a remote computer to another remote computer.

    :param transportsource: transport to be used for the source computer
    :param transportdestination: transport to be used for the destination computer
    :param str remotesource: path to the remote source directory / file
    :param str remotedestination: path to the remote destination directory / file
    :param kwargs: keyword parameters passed to the final put,
        except for 'dereference' that is passed to the initial get

    .. note:: it uses the method transportsource.copy_from_remote_to_remote
    """
    transportsource.copy_from_remote_to_remote(transportdestination,
                                               remotesource,remotedestination,
                                               **kwargs)

class Transport(object):
    """
    Abstract class for a generic transport (ssh, local, ...)
    Contains the set of minimal methods
    """
    # To be defined in the subclass
    # See the ssh or local plugin to see the format
    _valid_auth_params = None

    def __init__(self, *args, **kwargs):
        """
        __init__ method of the Transport base class.
        """
        self._logger = aiida.common.aiidalogger.getChild('transport').getChild(
            self.__class__.__name__)

        self._logger_extra = None

    def __enter__(self):
        """
        For transports that require opening a connection, opens
        all required channels (used in 'with' statements)
        """
        self.open()
        return self

    def __exit__(self, type, value, traceback):
        """
        Closes connections, if needed (used in 'with' statements).
        """
        self.close()

    def open(self):
        """
        Opens a local transport channel
        """
        raise NotImplementedError

    def close(self):
        """
        Closes the local transport channel
        """
        raise NotImplementedError

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, str(self))

    # redefine this in each subclass
    def __str__(self):
        return "[Transport class or subclass]"

    def _set_logger_extra(self, logger_extra):
        """
        Pass the data that should be passed automatically to self.logger
        as 'extra' keyword. This is typically useful if you pass data
        obtained using get_dblogger_extra in aiida.backends.djsite.utils, to automatically
        log also to the DbLog table.

        :param logger_extra: data that you want to pass as extra to the
          self.logger. To write to DbLog, it should be created by the
          aiida.backends.djsite.utils.get_dblogger_extra function. Pass None if you
          do not want to have extras passed.
        """
        self._logger_extra = logger_extra

    @classmethod
    def get_short_doc(self):
        """
        Return the first non-empty line of the class docstring, if available
        """
        # Remove empty lines
        docstring = self.__doc__
        if not docstring:
            return "No documentation available"

        doclines = [i for i in docstring.splitlines() if i.strip()]
        if doclines:
            return doclines[0].strip()
        else:
            return "No documentation available"

    @classmethod
    def get_valid_transports(cls):
        """
        :return: a list of existing plugin names
        """
        from aiida.common.pluginloader import existing_plugins

        return existing_plugins(Transport, "aiida.transport.plugins")

    @classmethod
    def get_valid_auth_params(cls):
        """
        Return the internal list of valid auth_params
        """
        if cls._valid_auth_params is None:
            raise NotImplementedError
        else:
            return cls._valid_auth_params

    @property
    def logger(self):
        """
        Return the internal logger.
        If you have set extra parameters using _set_logger_extra(), a
        suitable LoggerAdapter instance is created, bringing with itself
        also the extras.
        """
        try:
            import logging
            from aiida.utils.logger import get_dblogger_extra

            if self._logger_extra is not None:
                return logging.LoggerAdapter(logger=self._logger,
                                             extra=self._logger_extra)
            else:
                return self._logger
        except AttributeError:
            raise InternalError("No self._logger configured for {}!")

    def chdir(self, path):
        """
        Change directory to 'path'

        :param str path: path to change working directory into.
        :raises: IOError, if the requested path does not exist
        :rtype: string
        """
        # #TODO: understand if we want this behavior: this is emulated
        # by paramiko, and we should emulate it also for the local
        #        transport, since we do not want a global chdir for the whole
        #        code (the same holds for get_pwd).
        #        However, it could be useful to execute by default the
        #        codes from that specific directory.

        raise NotImplementedError


    def chmod(self, path, mode):
        """
        Change permissions of a path.

        :param str path: path to file
        :param int mode: new permissions
        """
        raise NotImplementedError


    def chown(self, path, uid, gid):
        """
        Change the owner (uid) and group (gid) of a file.
        As with python's os.chown function, you must pass both arguments,
        so if you only want to change one, use stat first to retrieve the
        current owner and group.

        :param str path: path to the file to change the owner and group of
        :param int uid: new owner's uid
        :param int gid: new group id
        """
        raise NotImplementedError


    def copy(self, remotesource, remotedestination, *args, **kwargs):
        """
        Copy a file or a directory from remote source to remote destination
        (On the same remote machine)

        :param str remotesource: path of the remote source directory / file
        :param str remotedestination: path of the remote destination directory / file

        :raises: IOError, if one of src or dst does not exist
        """
        raise NotImplementedError


    def copyfile(self, remotesource, remotedestination, *args, **kwargs):
        """
        Copy a file from remote source to remote destination
        (On the same remote machine)

        :param str remotesource: path of the remote source directory / file
        :param str remotedestination: path of the remote destination directory / file

        :raises IOError: if one of src or dst does not exist
        """
        raise NotImplementedError


    def copytree(self, remotesource, remotedestination, *args, **kwargs):
        """
        Copy a folder from remote source to remote destination
        (On the same remote machine)

        :param str remotesource: path of the remote source directory / file
        :param str remotedestination: path of the remote destination directory / file

        :raise IOError: if one of src or dst does not exist
        """
        raise NotImplementedError


    def copy_from_remote_to_remote(self,transportdestination,
                                      remotesource,remotedestination,**kwargs):
        """
        Copy files or folders from a remote computer to another remote computer.

        :param transportdestination: transport to be used for the destination computer
        :param str remotesource: path to the remote source directory / file
        :param str remotedestination: path to the remote destination directory / file
        :param kwargs: keyword parameters passed to the call to transportdestination.put,
            except for 'dereference' that is passed to self.get

        .. note:: the keyword 'dereference' SHOULD be set to False for the
         final put (onto the destination), while it can be set to the
         value given in kwargs for the get from the source. In that
         way, a symbolic link would never be followed in the final
         copy to the remote destination. That way we could avoid getting
         unknown (potentially malicious) files into the destination computer.
         HOWEVER, since dereference=False is currently NOT
         supported by all plugins, we still force it to True for the final put.

        .. note:: the supported keys in kwargs are callback, dereference,
           overwrite and ignore_nonexisting.
        """
        from aiida.common.folders import SandboxFolder

        kwargs_get = {'callback': None,
                      'dereference': kwargs.pop('dereference',True),
                      'overwrite': True,
                      'ignore_nonexisting': False,
                      }
        # TODO: dereference should be set to False in the following, as soon as
        # dereference=False is supported by all transport plugins
        kwargs_put = {'callback': kwargs.pop('callback',None),
                      'dereference': True,
                      'overwrite': kwargs.pop('overwrite',True),
                      'ignore_nonexisting': kwargs.pop('ignore_nonexisting',False),
                      }

        if kwargs:
            self.logger.error("Unknown parameters passed to copy_from_remote_to_remote")

        with SandboxFolder() as sandbox:
            self.get(remotesource, sandbox.abspath, **kwargs_get)
            # Then we scan the full sandbox directory with get_content_list,
            # because copying directly from sandbox.abspath would not work
            # to copy a single file into another single file, and copying
            # from sandbox.get_abs_path('*') would not work for files
            # beginning with a dot ('.').
            for filename in sandbox.get_content_list():
                transportdestination.put(os.path.join(sandbox.abspath,filename),
                                         remotedestination,**kwargs_put)


    def _exec_command_internal(self, command, **kwargs):
        """
        Execute the command on the shell, similarly to os.system.

        Enforce the execution to be run from the cwd (as given by
        self.getcwd), if this is not None.

        If possible, use the higher-level
        exec_command_wait function.

        :param str command: execute the command given as a string
        :return: stdin, stdout, stderr and the session, when this exists \
                 (can be None).
        """
        raise NotImplementedError


    def exec_command_wait(self, command, **kwargs):
        """
        Execute the command on the shell, waits for it to finish,
        and return the retcode, the stdout and the stderr.

        Enforce the execution to be run from the pwd (as given by
        self.getcwd), if this is not None.

        :param str command: execute the command given as a string
        :return: a list: the retcode (int), stdout (str) and stderr (str).
        """
        raise NotImplementedError


    def get(self, remotepath, localpath, *args, **kwargs):
        """
        Retrieve a file or folder from remote source to local destination
        dst must be an absolute path (src not necessarily)

        :param remotepath: (str) remote_folder_path
        :param localpath: (str) local_folder_path
        """
        raise NotImplementedError


    def getfile(self, remotepath, localpath, *args, **kwargs):
        """
        Retrieve a file from remote source to local destination
        dst must be an absolute path (src not necessarily)

        :param str remotepath: remote_folder_path
        :param str localpath: local_folder_path
        """
        raise NotImplementedError


    def gettree(self, remotepath, localpath, *args, **kwargs):
        """
        Retrieve a folder recursively from remote source to local destination
        dst must be an absolute path (src not necessarily)

        :param str remotepath: remote_folder_path
        :param str localpath: local_folder_path
        """
        raise NotImplementedError


    def getcwd(self):
        """
        Get working directory

        :return: a string identifying the current working directory
        """
        raise NotImplementedError


    def get_attribute(self, path):
        """
        Return an object FixedFieldsAttributeDict for file in a given path,
        as defined in aiida.common.extendeddicts
        Each attribute object consists in a dictionary with the following keys:

        * st_size: size of files, in bytes

        * st_uid: user id of owner

        * st_gid: group id of owner

        * st_mode: protection bits

        * st_atime: time of most recent access

        * st_mtime: time of most recent modification

        :param str path: path to file
        :return: object FixedFieldsAttributeDict
        """
        raise NotImplementedError


    def get_mode(self, path):
        """
        Return the portion of the file's mode that can be set by chmod().

        :param str path: path to file
        :return: the portion of the file's mode that can be set by chmod()
        """
        import stat

        return stat.S_IMODE(self.get_attribute(path).st_mode)


    def isdir(self, path):
        """
        True if path is an existing directory.

        :param str path: path to directory
        :return: boolean
        """
        raise NotImplementedError


    def isfile(self, path):
        """
        Return True if path is an existing file.

        :param str path: path to file
        :return: boolean
        """
        raise NotImplementedError


    def listdir(self, path='.', pattern=None):
        """
        Return a list of the names of the entries in the given path.
        The list is in arbitrary order. It does not include the special
        entries '.' and '..' even if they are present in the directory.

        :param str path: path to list (default to '.')
        :param str pattern: if used, listdir returns a list of files matching
                            filters in Unix style. Unix only.
        :return: a list of strings
        """
        raise NotImplementedError


    def makedirs(self, path, ignore_existing=False):
        """
        Super-mkdir; create a leaf directory and all intermediate ones.
        Works like mkdir, except that any intermediate path segment (not
        just the rightmost) will be created if it does not exist.

        :param str path: directory to create
        :param bool ignore_existing: if set to true, it doesn't give any error
                                     if the leaf directory does already exist

        :raises: OSError, if directory at path already exists
        """
        raise NotImplementedError


    def mkdir(self, path, ignore_existing=False):
        """
        Create a folder (directory) named path.

        :param str path: name of the folder to create
        :param bool ignore_existing: if True, does not give any error if the
                                     directory already exists

        :raises: OSError, if directory at path already exists
        """
        raise NotImplementedError


    def normalize(self, path='.'):
        """
        Return the normalized path (on the server) of a given path.
        This can be used to quickly resolve symbolic links or determine
        what the server is considering to be the "current folder".

        :param str path: path to be normalized

        :raise IOError: if the path can't be resolved on the server
        """
        raise NotImplementedError


    def put(self, localpath, remotepath, *args, **kwargs):
        """
        Put a file or a directory from local src to remote dst.
        src must be an absolute path (dst not necessarily))
        Redirects to putfile and puttree.

        :param str localpath: absolute path to local source
        :param str remotepath: path to remote destination
        """
        raise NotImplementedError


    def putfile(self, localpath, remotepath, *args, **kwargs):
        """
        Put a file from local src to remote dst.
        src must be an absolute path (dst not necessarily))

        :param str localpath: absolute path to local file
        :param str remotepath: path to remote file
        """
        raise NotImplementedError


    def puttree(self, localpath, remotepath, *args, **kwargs):
        """
        Put a folder recursively from local src to remote dst.
        src must be an absolute path (dst not necessarily))

        :param str localpath: absolute path to local folder
        :param str remotepath: path to remote folder
        """
        raise NotImplementedError


    def remove(self, path):
        """
        Remove the file at the given path. This only works on files;
        for removing folders (directories), use rmdir.

        :param str path: path to file to remove

        :raise IOError: if the path is a directory
        """
        raise NotImplementedError


    def rename(self, oldpath, newpath):
        """
        Rename a file or folder from oldpath to newpath.

        :param str oldpath: existing name of the file or folder
        :param str newpath: new name for the file or folder

        :raises IOError: if oldpath/newpath is not found
        :raises ValueError: if oldpath/newpath is not a valid string
        """
        raise NotImplementedError


    def rmdir(self, path):
        """
        Remove the folder named path.
        This works only for empty folders. For recursive remove, use rmtree.

        :param str path: absolute path to the folder to remove
        """
        raise NotImplementedError


    def rmtree(self, path):
        """
        Remove recursively the content at path

        :param str path: absolute path to remove
        """
        raise NotImplementedError

    def gotocomputer_command(self, remotedir):
        """
        Return a string to be run using os.system in order to connect
        via the transport to the remote directory.

        Expected behaviors:

        * A new bash session is opened

        * A reasonable error message is produced if the folder does not exist

        :param str remotedir: the full path of the remote directory
        """
        raise NotImplementedError

    def symlink(self, remotesource, remotedestination):
        """
        Create a symbolic link between the remote source and the remote
        destination.

        :param remotesource: remote source
        :param remotedestination: remote destination
        """
        raise NotImplementedError

    def whoami(self):
        """
        Get the remote username

        :return: list of username (str),
                 retval (int),
                 stderr (str)
        """
        # TODO : add tests for this method

        command = 'whoami'
        retval, username, stderr = self.exec_command_wait(command)
        if retval == 0:
            if stderr.strip():
                self.logger.warning("There was nonempty stderr in the whoami "
                                    "command: {}".format(stderr))
            return username.strip()
        else:
            self.logger.error("Problem executing whoami. Exit code: {}, stdout: '{}', "
                              "stderr: '{}'".format(retval, username, stderr))
            raise IOError("Error while executing whoami. Exit code: {}".format(retval))


    def path_exists(self, path):
        """
        Returns True if path exists, False otherwise.
        """
        raise NotImplementedError

    # The following definitions are almost copied and pasted
    # from the python module glob.
    def glob(self, pathname):
        """Return a list of paths matching a pathname pattern.

        The pattern may contain simple shell-style wildcards a la fnmatch.
        """
        return list(self.iglob(pathname))

    def iglob(self, pathname):
        """Return an iterator which yields the paths matching a pathname pattern.

        The pattern may contain simple shell-style wildcards a la fnmatch.

        """
        if not self.has_magic(pathname):
            # if os.path.lexists(pathname): # ORIGINAL
            # our implementation
            if self.path_exists(pathname):
                yield pathname
            return
        dirname, basename = os.path.split(pathname)
        if not dirname:
            # for name in self.glob1(os.curdir, basename): # ORIGINAL
            for name in self.glob1(self.getcwd(), basename):
                yield name
            return
        if self.has_magic(dirname):
            dirs = self.iglob(dirname)
        else:
            dirs = [dirname]
        if self.has_magic(basename):
            glob_in_dir = self.glob1
        else:
            glob_in_dir = self.glob0
        for dirname in dirs:
            for name in glob_in_dir(dirname, basename):
                yield os.path.join(dirname, name)

    # These 2 helper functions non-recursively glob inside a literal directory.
    # They return a list of basenames. `glob1` accepts a pattern while `glob0`
    # takes a literal basename (so it only has to check for its existence).

    def glob1(self, dirname, pattern):
        if not dirname:
            # dirname = os.curdir # ORIGINAL
            dirname = self.getcwd()
        if isinstance(pattern, unicode) and not isinstance(dirname, unicode):
            dirname = unicode(dirname, sys.getfilesystemencoding() or
                              sys.getdefaultencoding())
        try:
            # names = os.listdir(dirname)
            #print dirname
            names = self.listdir(dirname)
        except os.error:
            return []
        except IOError:
            return []
        if pattern[0] != '.':
            names = filter(lambda x: x[0] != '.', names)
        return fnmatch.filter(names, pattern)

    def glob0(self, dirname, basename):
        if basename == '':
            # `os.path.split()` returns an empty basename for paths ending with a
            # directory separator.  'q*x/' should match only directories.
            # if os.path.isdir(dirname):
            if self.isdir(dirname):
                return [basename]
        else:
            # if os.path.lexists(os.path.join(dirname, basename)):
            if self.path_exists(os.path.join(dirname, basename)):
                return [basename]
        return []

    def has_magic(self, s):
        return magic_check.search(s) is not None



# -*- coding: utf-8 -*-
"""
Test ssh plugin on localhost
"""
import unittest
import logging

import aiida.transport
import paramiko
from aiida.transport.plugins.ssh import SshTransport

# This will be used by test_all_plugins

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

plugin_transport = SshTransport(
    machine='localhost', timeout=30, 
    load_system_host_keys=True,
    key_policy = 'AutoAddPolicy')

class TestBasicConnection(unittest.TestCase):
    """
    Test basic connections.
    """
    
    def test_closed_connection_ssh(self):
        with self.assertRaises(aiida.transport.TransportInternalError):
            t = SshTransport(machine='localhost')
            t._exec_command_internal('ls')
            
    def test_closed_connection_sftp(self):
        with self.assertRaises(aiida.transport.TransportInternalError):
            t = SshTransport(machine='localhost')
            t.listdir()
            
    def test_invalid_param(self):
        with self.assertRaises(ValueError):
            SshTransport(machine='localhost', invalid_param=True)
                             

    def test_auto_add_policy(self):
        with SshTransport(machine='localhost', timeout=30, 
                          load_system_host_keys=True,
                          key_policy='AutoAddPolicy'):
            pass

    def test_no_host_key(self):
        # Disable logging to avoid output during test
        logging.disable(logging.ERROR)

        with self.assertRaises(paramiko.SSHException):
            with SshTransport(machine='localhost', timeout=30, 
                              load_system_host_keys = False):
                pass

        # Reset logging level
        logging.disable(logging.NOTSET)


if __name__ == '__main__': 
    unittest.main()



# -*- coding: utf-8 -*-

import json
import shutil
import tempfile
import os

from aiida.backends.djsite.db.testbase import AiidaTestCase
from aiida.common import utils
from aiida.common.additions.backup_script import backup_setup
from aiida.common.additions.backup_script.backup import Backup

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class TestBackupSetupScriptUnit(AiidaTestCase):

    def tearDown(self):
        utils.raw_input = None

    def test_construct_backup_variables(self):
        """
        Test that checks that the backup variables are populated as it
        should by the construct_backup_variables by asking the needed
        questions. A lambda function is used to simulate the user input.
        """
        _backup_setup_inst = backup_setup.BackupSetup()
        ac = utils.ArrayCounter()

        # Checking parsing of backup variables with many empty answers
        answers = ["", "y", "", "y", "", "y", "1", "y", "2", "y"]
        # utils.raw_input = lambda _: answers[self.array_counter()]
        utils.raw_input = lambda _: answers[ac.array_counter()]
        bk_vars = _backup_setup_inst.construct_backup_variables("")
        # Check the parsed answers
        self.assertIsNone(bk_vars[Backup._oldest_object_bk_key])
        self.assertIsNone(bk_vars[Backup._days_to_backup_key])
        self.assertIsNone(bk_vars[Backup._end_date_of_backup_key])
        self.assertEqual(bk_vars[Backup._periodicity_key], 1)
        self.assertEqual(bk_vars[Backup._backup_length_threshold_key], 2)

        # Checking parsing of backup variables with all the answers given
        ac = utils.ArrayCounter()
        answers = ["2013-07-28 20:48:53.197537+02:00", "y",
                    "2", "y", "2015-07-28 20:48:53.197537+02:00", "y",
                    "3", "y", "4", "y"]
        utils.raw_input = lambda _: answers[ac.array_counter()]
        bk_vars = _backup_setup_inst.construct_backup_variables("")
        # Check the parsed answers
        self.assertEqual(bk_vars[Backup._oldest_object_bk_key], answers[0])
        self.assertEqual(bk_vars[Backup._days_to_backup_key], 2)
        self.assertEqual(bk_vars[Backup._end_date_of_backup_key], answers[4])
        self.assertEqual(bk_vars[Backup._periodicity_key], 3)
        self.assertEqual(bk_vars[Backup._backup_length_threshold_key], 4)


class TestBackupSetupScriptIntegration(AiidaTestCase):

    def test_full_backup_setup_script(self):
        """
        This method is a full test of the backup setup script. It launches it,
        replies to all the question as the user would do and in the end it
        checks that the correct files were created with the right content.
        """
        # Create a temp folder where the backup files will be placed
        temp_folder = tempfile.mkdtemp()
        try:
            temp_aiida_folder = os.path.join(temp_folder, ".aiida")
            # The predefined answers for the setup script

            ac = utils.ArrayCounter()
            self.seq = -1
            answers = [temp_aiida_folder,   # the backup folder path
                       "",                  # should the folder be created?
                       "",                  # destination folder of the backup
                       "",                  # should the folder be created?
                       "n",                 # print config explanation?
                       "",                  # configure the backup conf file now?
                       "2014-07-18 13:54:53.688484+00:00", # start date of backup?
                       "",                  # is it correct?
                       "",                  # days to backup?
                       "",                  # is it correct?
                       "2015-04-11 13:55:53.688484+00:00", # end date of backup
                       "",                  # is it correct?
                       "1",                 # periodicity
                       "",                  # is it correct?
                       "2",                 # threshold?
                       ""]                  # is it correct?
            utils.raw_input = lambda _: answers[ac.array_counter()]

            # Run the setup script
            backup_setup.BackupSetup().run()

            # Get the backup configuration files & dirs
            backup_conf_records = [f for f in os.listdir(temp_aiida_folder)]
            # Check if all files & dirs are there
            self.assertTrue(backup_conf_records is not None and
                            len(backup_conf_records) == 4 and
                            "backup_dest" in backup_conf_records and
                            "backup_info.json.tmpl" in backup_conf_records and
                            "start_backup.py" in backup_conf_records and
                            "backup_info.json" in backup_conf_records,
                            "The created backup folder doesn't have the "
                            "expected files. "
                            "It contains: {}.".format(backup_conf_records))

            # Check the content of the main backup configuration file
            with open(os.path.join(temp_aiida_folder, "backup_info.json")
                      ) as conf_jfile:
                conf_cont = json.load(conf_jfile)
                self.assertEqual(conf_cont[Backup._oldest_object_bk_key],
                                 "2014-07-18 13:54:53.688484+00:00")
                self.assertEqual(conf_cont[Backup._days_to_backup_key], None)
                self.assertEqual(conf_cont[Backup._end_date_of_backup_key],
                                 "2015-04-11 13:55:53.688484+00:00")
                self.assertEqual(conf_cont[Backup._periodicity_key], 1)
                self.assertEqual(
                    conf_cont[Backup._backup_length_threshold_key], 2)
        finally:
            shutil.rmtree(temp_folder, ignore_errors=True)



#!/usr/bin/env runaiida
# -*- coding: utf-8 -*-

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

from aiida.backends.utils import load_dbenv, is_dbenv_loaded

if not is_dbenv_loaded():
    load_dbenv()

import sys
import os
from aiida.common.example_helpers import test_and_get_code
from aiida.orm import DataFactory
from aiida.common.exceptions import NotExistent
from aiida.orm.calculation.job.quantumespresso.pw import PwCalculation
from aiida.workflows2.run import asyncd

# If set to True, will ask AiiDA to run in serial mode (i.e., AiiDA will not
# invoke the mpirun command in the submission script)
run_in_serial_mode = False

################################################################

UpfData = DataFactory('upf')
ParameterData = DataFactory('parameter')
KpointsData = DataFactory('array.kpoints')
StructureData = DataFactory('structure')
try:
    dontsend = sys.argv[1]
    if dontsend == "--dont-send":
        submit_test = True
    elif dontsend == "--send":
        submit_test = False
    else:
        raise IndexError
except IndexError:
    print >> sys.stderr, ("The first parameter can only be either "
                          "--send or --dont-send")
    sys.exit(1)

try:
    codename = sys.argv[2]
except IndexError:
    codename = None

# If True, load the pseudos from the family specified below
# Otherwise, use static files provided
auto_pseudos = False

queue = None
# queue = "Q_aries_free"
#####

alat = 4.  # angstrom
cell = [[alat, 0., 0., ],
        [0., alat, 0., ],
        [0., 0., alat, ],
        ]

# BaTiO3 cubic structure
s = StructureData(cell=cell)
s.append_atom(position=(0., 0., 0.), symbols=['Ba'])
s.append_atom(position=(alat / 2., alat / 2., alat / 2.), symbols=['Ti'])
s.append_atom(position=(alat / 2., alat / 2., 0.), symbols=['O'])
s.append_atom(position=(alat / 2., 0., alat / 2.), symbols=['O'])
s.append_atom(position=(0., alat / 2., alat / 2.), symbols=['O'])

elements = list(s.get_symbols_set())

if auto_pseudos:
    valid_pseudo_groups = UpfData.get_upf_groups(filter_elements=elements)

    try:
        pseudo_family = sys.argv[3]
    except IndexError:
        print >> sys.stderr, "Error, auto_pseudos set to True. You therefore need to pass as second parameter"
        print >> sys.stderr, "the pseudo family name."
        print >> sys.stderr, "Valid UPF families are:"
        print >> sys.stderr, "\n".join("* {}".format(i.name) for i in valid_pseudo_groups)
        sys.exit(1)

    try:
        UpfData.get_upf_group(pseudo_family)
    except NotExistent:
        print >> sys.stderr, "auto_pseudos is set to True and pseudo_family='{}',".format(pseudo_family)
        print >> sys.stderr, "but no group with such a name found in the DB."
        print >> sys.stderr, "Valid UPF groups are:"
        print >> sys.stderr, ",".join(i.name for i in valid_pseudo_groups)
        sys.exit(1)

parameters = ParameterData(dict={
    'CONTROL': {
        'calculation': 'scf',
        'restart_mode': 'from_scratch',
        'wf_collect': True,
        'tstress': True,
        'tprnfor': True,
    },
    'SYSTEM': {
        'ecutwfc': 40.,
        'ecutrho': 320.,
    },
    'ELECTRONS': {
        'conv_thr': 1.e-10,
    }
})

# Set up kpoint
kpoints = KpointsData()
kpoints_mesh = 2
kpoints.set_kpoints_mesh([kpoints_mesh, kpoints_mesh, kpoints_mesh])

# to retrieve the bands
# (the object settings is optional)
settings_dict = {'also_bands': True}
settings = ParameterData(dict=settings_dict)

# calc.label = "Test QE pw.x"
# calc.description = "Test calculation with the Quantum ESPRESSO pw.x code"
# Valid only for Slurm and PBS (using default values for the
# number_cpus_per_machine), change for SGE-like schedulers


JobCalc = PwCalculation.process()

# Inputs
inputs = JobCalc.get_inputs_template()
inputs.structure = s
inputs.parameters= parameters
inputs.kpoints = kpoints
inputs.code=test_and_get_code(codename, expected_code_type='quantumespresso.pw')

# Calculation options
options = inputs._options
options.max_wallclock_seconds = 30 * 60  # 30 min
options.resources = {"num_machines": 1, "num_mpiprocs_per_machine": 8}
options.custom_scheduler_commands = u"#SBATCH --account=ch3"
if run_in_serial_mode:
    options.withmpi = False
if queue is not None:
    options.queue_name = queue

if settings is not None:
    inputs.settings = settings

    # if auto_pseudos:
    #     try:
    #         calc.use_pseudos_from_family(pseudo_family)
    #         print "Pseudos successfully loaded from family {}".format(pseudo_family)
    #     except NotExistent:
    #         print ("Pseudo or pseudo family not found. You may want to load the "
    #                "pseudo family, or set auto_pseudos to False.")
    #         raise
    # else:
    raw_pseudos = [
        ("Ba.pbesol-spn-rrkjus_psl.0.2.3-tot-pslib030.UPF", 'Ba', 'pbesol'),
        ("Ti.pbesol-spn-rrkjus_psl.0.2.3-tot-pslib030.UPF", 'Ti', 'pbesol'),
        ("O.pbesol-n-rrkjus_psl.0.1-tested-pslib030.UPF", 'O', 'pbesol')]

    pseudos_to_use = {}
    for fname, elem, pot_type in raw_pseudos:
        absname = os.path.realpath(
            os.path.join(
                os.path.dirname(__file__), "..", "submission", "data", fname))
        pseudo, created = UpfData.get_or_create(absname, use_first=True)
        if created:
            print "Created the pseudo for {}".format(elem)
        else:
            print "Using the pseudo for {} from DB: {}".format(elem, pseudo.pk)
        pseudos_to_use[elem] = pseudo
    inputs.pseudo = pseudos_to_use

# Run the calculation
print(asyncd(JobCalc, **inputs))
